--------------------------------------------------------------------------------
-- Copyright (c) 2000 by Trenz Electronic.
-- Duenner Kirchweg 77, 32257 Buende, Germany, www.trenz-electronic.de
--     
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--     
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--     
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
--------------------------------------------------------------------------------
-- Project:      MCS51 Compatible 8-bit Microcontroller Core
-- File:         TE51eval.vhd
-- Description:  Behavioral Evaluation Model.
-- Version:      FB, 2000nov07 - beta
--------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.all;

package te51pak is
	constant te51alias_SP:  STD_LOGIC_VECTOR(7 downto 0):= x"81";
	constant te51alias_DPL: STD_LOGIC_VECTOR(7 downto 0):= x"82";
	constant te51alias_DPH: STD_LOGIC_VECTOR(7 downto 0):= x"83";
	constant te51alias_PSW: STD_LOGIC_VECTOR(7 downto 0):= x"D0";
	constant te51alias_CY:  STD_LOGIC_VECTOR(7 downto 0):= x"D7";
	constant te51alias_A:   STD_LOGIC_VECTOR(7 downto 0):= x"E0";
	----------------------------------------------------------------------------
	type te51addrT is (
	te51addr_none,
	te51addr_A,            -- Akkumulator
	te51addr_datain,       -- datain port
	te51addr_DPL,          -- DPTR lo
	te51addr_DPH,          -- DPTR hi
	te51addr_direct,       -- direct address (opcode byte 2)
	te51addr_direct3,      -- direct address (opcode byte 3)
	te51addr_bit,          -- bit address
	te51addr_xdata2,       -- #data, opcode byte 2
	te51addr_xdata3,       -- #data, opcode byte 3
	te51addr_Rn,           -- Rn
	te51addr_aRi,          -- @Ri
	te51addr_cy16,         -- carry from last slot
	te51addr_c0,           -- constant #0
	te51addr_c1,           -- constant #1
	te51addr_const2,       -- constant #2
	te51addr_PC_rel2,      -- Program Counter relative, opcode byte2
	te51addr_PC_rel3,      -- Program Counter relative, opcode byte3
	te51addr_PC_addr11,    -- Program Counter, 11bit absolute
	te51addr_PC_addr16,    -- Program Counter, 16bit absolute
	te51addr_SP,           -- Stack pointer
	te51addr_aSP,          -- Stack indirect
	te51addr_ne,           -- not equal result
	te51addr_jeq,          -- jump if equal
	te51addr_jc,           -- jump if carry set
	te51addr_jnc,          -- jump if carry not set
	te51addr_jmp,          -- jump (without condition)
	te51addr_call,         -- call (without condition)
	te51addr_ret,          -- return (without condition)
	te51addr_CY,           -- carry flag
	te51addr_dummy
	);
	
	----------------------------------------------------------------------------
	type te51opT is	(
	te51op_none,
	te51op_add,                                 -- add
	te51op_addc,                                -- add with carry
	te51op_sub,                                 -- sub
	te51op_subb,                                -- subb with borrow
	te51op_src2,                                -- src #2
	te51op_cmp,                                 -- compare
	te51op_clt,                                 -- set carry when "less than"
	te51op_and,                                 -- bitwise and
	te51op_or,                                  -- bitwise or
	te51op_swap                                 -- swap nibbles
	);
	----------------------------------------------------------------------------
	type te51mcodeT is record
		-- general instruction info
		bytes:              INTEGER range 1 to 3; -- # of bytes
		cycles:             INTEGER range 1 to 4; -- # of machine cycles
		movxrd:             STD_LOGIC;            -- '1' for MOVX-type reads
		movxwr:             STD_LOGIC;            -- '1' for MOVX-type writes
		-- phase 1: C1S1P1 .. C1S3P2
		--		op1:                te51opT;
		--		adx11, adx12:       te51addrT;
		-- phase 2: C2S4P1 .. C1S6P2
		op2:                te51opT;
		adx21, adx22:       te51addrT;
		rslt2, psw2:        STD_LOGIC;
		-- phase 3: C2S1P1 .. C2S3P2
		op3:                te51opT;
		adx31, adx32:       te51addrT;
		rslt3, psw3:        STD_LOGIC;
		-- phase 4: C2S4P1 .. C2S6P2
	end record;
	
	----------------------------------------------------------------------------
	type te51stateT is (
	C1S1P1, C1S1P2, C1S2P1, C1S2P2, C1S3P1, C1S3P2, C1S4P1, C1S4P2, C1S5P1, C1S5P2, C1S6P1, C1S6P2,
	C2S1P1, C2S1P2, C2S2P1, C2S2P2, C2S3P1, C2S3P2, C2S4P1, C2S4P2, C2S5P1, C2S5P2, C2S6P1, C2S6P2,
	C3S1P1, C3S1P2, C3S2P1, C3S2P2, C3S3P1, C3S3P2, C3S4P1, C3S4P2, C3S5P1, C3S5P2, C3S6P1, C3S6P2,
	C4S1P1, C4S1P2, C4S2P1, C4S2P2, C4S3P1, C4S3P2, C4S4P1, C4S4P2, C4S5P1, C4S5P2, C4S6P1, C4S6P2
	);
	
	----------------------------------------------------------------------------
	type te51instrT is (	
	te51_reserved,
	-- Arithmetic Operations
	te51_ADD_A_Rn,           -- ADD   A, Rn           Add register to Accumulator
	te51_ADD_A_direct,       -- ADD   A, direct       Add direct byte to Accumulator
	te51_ADD_A_aRi,          -- ADD   A, @Ri          Add indirect RAM to Accumulator
	te51_ADD_A_xdata,        -- ADD   A, #data        Add immediate data to Accumulator
	te51_ADDC_A_Rn,          -- ADDC  A, Rn           Add register to Accumulator with Carry
	te51_ADDC_A_direct,      -- ADDC  A, direct       Add direct byte to Accumulator with Carry
	te51_ADDC_A_aRi,         -- ADDC  A, @Ri          Add indirect RAM to Accumulator with Carry
	te51_ADDC_A_xdata,       -- ADDC  A, #data        Add immediate data to Acc with Carry
	te51_SUBB_A_Rn,          -- SUBB  A, Rn           Subtract Register from Acc with borryw
	te51_SUBB_A_direct,      -- SUBB  A, direct       Subtract direct byte from Acc with borrow
	te51_SUBB_A_aRi,         -- SUBB  A, @Ri          Subtract indirect RAM from ACC with borrow
	te51_SUBB_A_xdata,       -- SUBB  A, #data        Subtract immediate data from Acc with borrow
	te51_INC_A,              -- INC   A               Increment Accumulator
	te51_INC_Rn,             -- INC   Rn              Increment register
	te51_INC_direct,         -- INC   direct          Increment direct byte
	te51_INC_aRi,            -- INC   @Ri             Increment direct RAM
	te51_DEC_A,              -- DEC   A               Decrement Accumulator
	te51_DEC_Rn,             -- DEC   Rn              Decrement Register
	te51_DEC_direct,         -- DEC   direct          Decrement direct byte
	te51_DEC_aRi,            -- DEC   @Ri             Decrement indirect RAM
	te51_INC_DPTR,           -- INC   DPTR            Increment Data Pointer
	te51_MUL_AB,             -- MUL   AB              Multiply A & B
	te51_DIV_AB,             -- DIV   AB              Divide A by B
	te51_DA_A,               -- DA    A               Decimal Adjust Accumulator
	-- Logical Operations
	te51_ANL_A_Rn,           -- ANL   A, Rn           AND Register to Accumulator
	te51_ANL_A_direct,       -- ANL   A, direct       AND direct byte to Accumulator
	te51_ANL_A_aRi,          -- ANL   A, @Ri          AND indirect RAM to Accumulator
	te51_ANL_A_xdata,        -- ANL   A, #data        AND immediate data to Accumulator
	te51_ANL_direct_A,       -- ANL   direct, A       AND Accumulator to direct byte
	te51_ANL_direct_xdata,   -- ANL   direct, #data   AND immediate data to direct byte
	te51_ORL_A_Rn,           -- ORL   A, Rn           OR register to Accumulator
	te51_ORL_A_direct,       -- ORL   A, direct       OR direct byte to Accumulator
	te51_ORL_A_aRi,          -- ORL   A, @Ri          OR indirect RAM to Accumulator
	te51_ORL_A_xdata,        -- ORL   A, #data        OR immediate data to Accumulator
	te51_ORL_direct_A,       -- ORL   direct, A       OR Accumulator to direct byte
	te51_ORL_direct_xdata,   -- ORL   direct, #data   OR immediate data to direct byte
	te51_XRL_A_Rn,           -- XRL   A, Rn           Exclusive-OR register to Accumulator
	te51_XRL_A_direct,       -- XRL   A, direct       Exclusive-OR direct byte to Accumulator
	te51_XRL_A_aRi,          -- XRL   A, @Ri          Exclusive-OR indirect RAM to Accumulator
	te51_XRL_A_xdata,        -- XRL   A, #data        Exclusive-OR immediate data to Accumulator
	te51_XRL_direct_A,       -- XRL   direct, A       Exclusive-OR Accumulator to direct byte
	te51_XRL_direct_xdata,   -- XRL   direct, #data   Exclusive-OR immediate data to direct byte
	te51_CLR_A,              -- CLR   A               Clear Accumulator
	te51_CPL_A,              -- CPL   A               Complement Accumulator
	te51_RL_A,               -- RL    A               Rotate Accumulator Left
	te51_RLC_A,              -- RLC   A               Rotate Accumulator Left through the Carry
	te51_RR_A,               -- RR    A               Rotate Accumulator Right
	te51_RRC_A,              -- RRC   A               Rotate Accumulator Right through the Carry
	te51_SWAP_A,             -- SWAP  A               Swap nibbles within the Accumulator
	-- Data transfer
	te51_MOV_A_Rn,           -- MOV   A, Rn           Move register to Accumulator
	te51_MOV_A_direct,       -- MOV   A, direct       Move direct byte to Accumulator
	te51_MOV_A_aRi,          -- MOV   A, @Ri          Move indirect RAM to Accumulator
	te51_MOV_A_xdata,        -- MOV   A, #data        Move immediate data to Accumulator
	te51_MOV_Rn_A,           -- MOV   Rn, A           Move Accumulator to register
	te51_MOV_Rn_direct,      -- MOV   Rn, direct      Move direct byte to register
	te51_MOV_Rn_xdata,       -- MOV   Rn, #data       Move immediate data to register
	te51_MOV_direct_A,       -- MOV   direct, A       Move Accumulator to direct byte
	te51_MOV_direct_Rn,      -- MOV   direct, Rn      Move register to direct byte
	te51_MOV_direct_direct,  -- MOV   direct, direct  Move direct byte to direct
	te51_MOV_direct_aRi,     -- MOV   direct, @Ri     Move indirect RAM to direct byte
	te51_MOV_direct_xdata,   -- MOV   direct, #data   Move immediate data to direct byte
	te51_MOV_aRi_A,          -- MOV   @Ri, A          Move Accumulator to indirect RAM
	te51_MOV_aRi_direct,     -- MOV   @Ri, direct     Move direct byte to indirect RAM
	te51_MOV_aRi_xdata,      -- MOV   @Ri, #data      Move immediate data to indirect RAM
	te51_MOV_DPTR_xdata16,   -- MOV   DPTR, #data16   Load Data Pointer with a 16-bit constant
	te51_MOVC_A_aA_DPTR,     -- MOVC  A, @A+DPTR      Move Code byte relative to DPTR to Acc
	te51_MOVC_A_aA_PC,       -- MOVC  A, @A+PC        Move Code byte relative to PC to Acc
	te51_MOVX_A_aRi,         -- MOVX  A, @Ri          Move External RAM (8-bit addr) to Acc
	te51_MOVX_A_aDPTR,       -- MOVX  A, @DPTR        Move External RAM (16-bit addr) to Acc
	te51_MOVX_aRi_A,         -- MOVX  @Ri, A          Move Acc to External RAM (8-bit addr)
	te51_MOVX_aDPTR_A,       -- MOVX  @DPTR, A        Move Acc to External RAM (16-bit addr)
	te51_PUSH_direct,        -- PUSH  direct          Push direct byte onto stack
	te51_POP_direct,         -- POP   direct          Pop direct byte from stack
	te51_XCH_A_Rn,           -- XCH   A, Rn           Exchange register with Accumulator
	te51_XCH_A_direct,       -- XCH   A, direct       Exchange direct byte with Accumulator
	te51_XCH_A_aRi,          -- XCH   A, @Ri          Exchange indirect RAM with Accumulator
	te51_XCHD_A_aRi,         -- XCHD  A, @Ri          Exchange low-order Digit indirect RAM with Acc
	-- Boolean variable manipulation
	te51_CLR_C,              -- CLR   C               Clear Carry
	te51_CLR_bit,            -- CLR   bit             Clear direct bit
	te51_SETB_C,             -- SETB  C               Set Carry
	te51_SETB_bit,           -- SETB  bit             Set direct bit
	te51_CPL_C,              -- CPL   C               Complement Carry
	te51_CPL_bit,            -- CPL   bit             Complement direct bit
	te51_ANL_C_bit,          -- ANL   C, bit          AND direct bit to CARRY
	te51_ANL_C_nbit,         -- ANL   C, /bit         AND complement of direct bit to Carry
	te51_ORL_C_bit,          -- ORL   C, bit          OR direct bit to Carry
	te51_ORL_C_nbit,         -- ORL   C, /bit         OR complement of direct bit to Carry
	te51_MOV_C_bit,          -- MOV   C, bit          Move direct bit to Carry
	te51_MOV_bit_C,          -- MOV   bit, C          Move Carry to direct bit
	te51_JC_rel,             -- JC    rel             Jump if Carry is set
	te51_JNC_rel,            -- JNZ   rel             Jump if Carry not set
	te51_JB_bit_rel,         -- JB    bit, rel        Jump if direct Bit is set
	te51_JNB_bit_rel,        -- JNB   bit, rel        Jump if direct Bit is Not set
	te51_JBC_bit_rel,        -- JBC   bit, rel        Jump if direct Bit si set & clear bit
	-- Program branching
	te51_ACALL_addr11,       -- ACALL addr11          Absolute Subroutine Call
	te51_LCALL_addr16,       -- LCALL addr16          Long subroutine Call
	te51_RET,                -- RET                   Return from Subroutine
	te51_RETI,               -- RETI                  Return from interrupt
	te51_AJMP_addr11,        -- AJMP  addr11          Absolute Jump
	te51_LJMP_addr16,        -- LJMP  addr16          Long Jump
	te51_SJMP_rel,           -- SJMP  rel             Short Jump (relative addr)
	te51_JMP_aA_DPTR,        -- JMP   @A+DPTR         Jump indirect relative to the DPTR
	te51_JZ_rel,             -- JZ    rel             Jump if Accumulator is Zero
	te51_JNZ_rel,            -- JNZ   rel             Jump if Accumulator is Not Zero
	te51_CJNE_A_direct_rel,  -- CJNE  A, direct, rel  Compare direct byte to Acc and Jump if Not Equal
	te51_CJNE_A_xdata_rel,   -- CJNE  A, #data, rel   Compare immediate to Acc and Jump if Not Equal
	te51_CJNE_Rn_xdata_rel,  -- CJNE  Rn, #data, rel  Compare immediate to register and Jump if Not Equal
	te51_CJNE_aRi_xdata_rel, -- CJNE  @Ri, #data, rel Compare immediate to indirect and Jump if Not Equal
	te51_DJNZ_Rn_rel,        -- DJNZ  Rn, rel         Decrement register and Jump if Not Zero
	te51_DJNZ_direct_rel,    -- DJNZ  direct, rel     Decrement direct byte and Jump if Not Zero
	te51_NOP                 -- NOP                   No Operation
	);
	
end package;
use work.te51pak.all;

library IEEE;
use IEEE.STD_LOGIC_1164.all;

entity te51dec is
	port(
		opcode1 : in STD_LOGIC_VECTOR(7 downto 0);
		instr:   out te51instrT
		);
end te51dec;

--------------------------------------------------------------------------------
architecture bhv of te51dec is
begin
	
	with opcode1 select instr<=
	te51_NOP                when x"00", -- NOP
	te51_ACALL_addr11       when x"01", -- AJMP  code addr
	te51_LJMP_addr16        when x"02", -- LJMP  code addr
	te51_RR_A               when x"03", -- RR    A
	te51_INC_A              when x"04", -- INC   A
	te51_INC_direct         when x"05", -- INC   data addr
	te51_INC_aRi            when x"06", -- INC   @R0
	te51_INC_aRi            when x"07", -- INC   @R1
	te51_INC_Rn             when x"08", -- INC   R0
	te51_INC_Rn             when x"09", -- INC   R1
	te51_INC_Rn             when x"0A", -- INC   R2
	te51_INC_Rn             when x"0B", -- INC   R3
	te51_INC_Rn             when x"0C", -- INC   R4
	te51_INC_Rn             when x"0D", -- INC   R5
	te51_INC_Rn             when x"0E", -- INC   R6
	te51_INC_Rn             when x"0F", -- INC   R7
	te51_JBC_bit_rel        when x"10", -- JBC   bit addr, code addr
	te51_ACALL_addr11       when x"11", -- ACALL code addr
	te51_LCALL_addr16       when x"12", -- LCALL code addr
	te51_RRC_A              when x"13", -- RRC   A
	te51_DEC_A              when x"14", -- DEC   A
	te51_DEC_direct         when x"15", -- DEC   data addr
	te51_DEC_aRi            when x"16", -- DEC   @R0
	te51_DEC_aRi            when x"17", -- DEC   @R1
	te51_DEC_Rn             when x"18", -- DEC   R0
	te51_DEC_Rn             when x"19", -- DEC   R1
	te51_DEC_Rn             when x"1A", -- DEC   R2
	te51_DEC_Rn             when x"1B", -- DEC   R3
	te51_DEC_Rn             when x"1C", -- DEC   R4
	te51_DEC_Rn             when x"1D", -- DEC   R5
	te51_DEC_Rn             when x"1E", -- DEC   R6
	te51_DEC_Rn             when x"1F", -- DEC   R7
	te51_JB_bit_rel         when x"20", -- JB    bit addr, code addr
	te51_AJMP_addr11        when x"21", -- AJMP  code addr
	te51_RET                when x"22", -- RET
	te51_RL_A               when x"23", -- RL    A
	te51_ADD_A_xdata        when x"24", -- ADD   A, #data
	te51_ADD_A_direct       when x"25", -- ADD   A, data addr
	te51_ADD_A_aRi          when x"26", -- ADD   A, @R0
	te51_ADD_A_aRi          when x"27", -- ADD   A, @R1
	te51_ADD_A_Rn           when x"28", -- ADD   A, R0
	te51_ADD_A_Rn           when x"29", -- ADD   A, R1
	te51_ADD_A_Rn           when x"2A", -- ADD   A, R2
	te51_ADD_A_Rn           when x"2B", -- ADD   A, R3
	te51_ADD_A_Rn           when x"2C", -- ADD   A, R4
	te51_ADD_A_Rn           when x"2D", -- ADD   A, R5
	te51_ADD_A_Rn           when x"2E", -- ADD   A, R6
	te51_ADD_A_Rn           when x"2F", -- ADD   A, R7
	te51_JNB_bit_rel        when x"30", -- JNB   bit addr, code addr
	te51_ACALL_addr11       when x"31", -- ACALL code addr
	te51_RETI               when x"32", -- RETI
	te51_RLC_A              when x"33", -- RLC   A
	te51_ADDC_A_xdata       when x"34", -- ADDC  A, #data
	te51_ADDC_A_direct      when x"35", -- ADDC  A, data addr
	te51_ADDC_A_aRi         when x"36", -- ADDC  A, @R0
	te51_ADDC_A_aRi         when x"37", -- ADDC  A, @R1
	te51_ADDC_A_Rn          when x"38", -- ADDC  A, R0
	te51_ADDC_A_Rn          when x"39", -- ADDC  A, R1
	te51_ADDC_A_Rn          when x"3A", -- ADDC  A, R2
	te51_ADDC_A_Rn          when x"3B", -- ADDC  A, R3
	te51_ADDC_A_Rn          when x"3C", -- ADDC  A, R4
	te51_ADDC_A_Rn          when x"3D", -- ADDC  A, R5
	te51_ADDC_A_Rn          when x"3E", -- ADDC  A, R6
	te51_ADDC_A_Rn          when x"3F", -- ADDC  A, R7
	te51_JC_rel             when x"40", -- JC    code addr
	te51_AJMP_addr11        when x"41", -- AJMP  code addr
	te51_ORL_direct_A       when x"42", -- ORL   data addr, A
	te51_ORL_direct_xdata   when x"43", -- ORL   data addr, #data
	te51_ORL_A_xdata        when x"44", -- ORL   A, #data
	te51_ORL_A_direct       when x"45", -- ORL   A, data addr
	te51_ORL_A_aRi          when x"46", -- ORL   A, @R0
	te51_ORL_A_aRi          when x"47", -- ORL   A, @R1
	te51_ORL_A_Rn           when x"48", -- ORL   A, R0
	te51_ORL_A_Rn           when x"49", -- ORL   A, R1
	te51_ORL_A_Rn           when x"4A", -- ORL   A, R2
	te51_ORL_A_Rn           when x"4B", -- ORL   A, R3
	te51_ORL_A_Rn           when x"4C", -- ORL   A, R4
	te51_ORL_A_Rn           when x"4D", -- ORL   A, R5
	te51_ORL_A_Rn           when x"4E", -- ORL   A, R6
	te51_ORL_A_Rn           when x"4F", -- ORL   A, R7
	te51_JNC_rel            when x"50", -- JNC   code addr
	te51_ACALL_addr11       when x"51", -- ACALL code addr
	te51_ANL_direct_A       when x"52", -- ANL   data addr, A
	te51_ANL_direct_xdata   when x"53", -- ANL   data addr, #data
	te51_ANL_A_xdata        when x"54", -- ANL   A, #data
	te51_ANL_A_direct       when x"55", -- ANL   A, data addr
	te51_ANL_A_aRi          when x"56", -- ANL   A, @R0
	te51_ANL_A_aRi          when x"57", -- ANL   A, @R1
	te51_ANL_A_Rn           when x"58", -- ANL   A, R0
	te51_ANL_A_Rn           when x"59", -- ANL   A, R1
	te51_ANL_A_Rn           when x"5A", -- ANL   A, R2
	te51_ANL_A_Rn           when x"5B", -- ANL   A, R3
	te51_ANL_A_Rn           when x"5C", -- ANL   A, R4
	te51_ANL_A_Rn           when x"5D", -- ANL   A, R5
	te51_ANL_A_Rn           when x"5E", -- ANL   A, R6
	te51_ANL_A_Rn           when x"5F", -- ANL   A, R7
	te51_JZ_rel             when x"60", -- JZ    code addr
	te51_AJMP_addr11        when x"61", -- AJMP  code addr
	te51_XRL_direct_A       when x"62", -- XRL   data addr, A
	te51_XRL_direct_xdata   when x"63", -- XRL   data addr, #data
	te51_XRL_A_xdata        when x"64", -- XRL   A, #data
	te51_XRL_A_direct       when x"65", -- XRL   A, data addr
	te51_XRL_A_aRi          when x"66", -- XRL   A, @R0
	te51_XRL_A_aRi          when x"67", -- XRL   A, @R1
	te51_XRL_A_Rn           when x"68", -- XRL   A, R0
	te51_XRL_A_Rn           when x"69", -- XRL   A, R1
	te51_XRL_A_Rn           when x"6A", -- XRL   A, R2
	te51_XRL_A_Rn           when x"6B", -- XRL   A, R3
	te51_XRL_A_Rn           when x"6C", -- XRL   A, R4
	te51_XRL_A_Rn           when x"6D", -- XRL   A, R5
	te51_XRL_A_Rn           when x"6E", -- XRL   A, R6
	te51_XRL_A_Rn           when x"6F", -- XRL   A, R7
	te51_JNZ_rel            when x"70", -- JNZ   code addr
	te51_ACALL_addr11       when x"71", -- ACALL code addr
	te51_ORL_C_bit          when x"72", -- ORL   C, bit addr
	te51_JMP_aA_DPTR        when x"73", -- JMP   @A+DPTR
	te51_MOV_A_xdata        when x"74", -- MOV   A, #data
	te51_MOV_direct_xdata   when x"75", -- MOV   data addr, #data
	te51_MOV_aRi_xdata      when x"76", -- MOV   @R0, #data
	te51_MOV_aRi_xdata      when x"77", -- MOV   @R1, #data
	te51_MOV_Rn_xdata       when x"78", -- MOV   R0, #data
	te51_MOV_Rn_xdata       when x"79", -- MOV   R1, #data
	te51_MOV_Rn_xdata       when x"7A", -- MOV   R2, #data
	te51_MOV_Rn_xdata       when x"7B", -- MOV   R3, #data
	te51_MOV_Rn_xdata       when x"7C", -- MOV   R4, #data
	te51_MOV_Rn_xdata       when x"7D", -- MOV   R5, #data
	te51_MOV_Rn_xdata       when x"7E", -- MOV   R6, #data
	te51_MOV_Rn_xdata       when x"7F", -- MOV   R7, #data
	te51_SJMP_rel           when x"80", -- SJMP  code addr
	te51_AJMP_addr11        when x"81", -- AJMP  code addr
	te51_ANL_C_bit          when x"82", -- ANL   C, bit addr
	te51_MOVC_A_aA_PC       when x"83", -- MOVC  A, @A+PC
	te51_DIV_AB             when x"84", -- DIV   AB
	te51_MOV_direct_direct  when x"85", -- MOV   data addr, data addr
	te51_MOV_direct_aRi     when x"86", -- MOV   data addr, @R0
	te51_MOV_direct_aRi     when x"87", -- MOV   data addr, @R1
	te51_MOV_direct_Rn      when x"88", -- MOV   data addr, R0
	te51_MOV_direct_Rn      when x"89", -- MOV   data addr, R1
	te51_MOV_direct_Rn      when x"8A", -- MOV   data addr, R2
	te51_MOV_direct_Rn      when x"8B", -- MOV   data addr, R3
	te51_MOV_direct_Rn      when x"8C", -- MOV   data addr, R4
	te51_MOV_direct_Rn      when x"8D", -- MOV   data addr, R5
	te51_MOV_direct_Rn      when x"8E", -- MOV   data addr, R6
	te51_MOV_direct_Rn      when x"8F", -- MOV   data addr, R7
	te51_MOV_DPTR_xdata16   when x"90", -- MOV   DPTR, #data
	te51_ACALL_addr11       when x"91", -- ACALL code addr
	te51_MOV_bit_C          when x"92", -- MOV   bit addr, C
	te51_MOVC_A_aA_DPTR     when x"93", -- MOVC  A, @A+DPTR
	te51_SUBB_A_xdata       when x"94", -- SUBB  A, #data
	te51_SUBB_A_direct      when x"95", -- SUBB  A, data addr
	te51_SUBB_A_aRi         when x"96", -- SUBB  A, @R0
	te51_SUBB_A_aRi         when x"97", -- SUBB  A, @R1
	te51_SUBB_A_Rn          when x"98", -- SUBB  A, R0
	te51_SUBB_A_Rn          when x"99", -- SUBB  A, R1
	te51_SUBB_A_Rn          when x"9A", -- SUBB  A, R2
	te51_SUBB_A_Rn          when x"9B", -- SUBB  A, R3
	te51_SUBB_A_Rn          when x"9C", -- SUBB  A, R4
	te51_SUBB_A_Rn          when x"9D", -- SUBB  A, R5
	te51_SUBB_A_Rn          when x"9E", -- SUBB  A, R6
	te51_SUBB_A_Rn          when x"9F", -- SUBB  A, R7
	te51_ORL_C_nbit         when x"A0", -- ORL   C, /bit addr
	te51_AJMP_addr11        when x"A1", -- AJMP  code addr
	te51_MOV_C_bit          when x"A2", -- MOV   C, bit addr
	te51_INC_DPTR           when x"A3", -- INC   DPTR
	te51_MUL_AB             when x"A4", -- MUL   AB
	te51_reserved           when x"A5", -- reserved
	te51_MOV_aRi_direct     when x"A6", -- MOV   @R0, data addr
	te51_MOV_aRi_direct     when x"A7", -- MOV   @R1, data addr
	te51_MOV_Rn_direct      when x"A8", -- MOV   R0, data addr
	te51_MOV_Rn_direct      when x"A9", -- MOV   R1, data addr
	te51_MOV_Rn_direct      when x"AA", -- MOV   R2, data addr
	te51_MOV_Rn_direct      when x"AB", -- MOV   R3, data addr
	te51_MOV_Rn_direct      when x"AC", -- MOV   R4, data addr
	te51_MOV_Rn_direct      when x"AD", -- MOV   R5, data addr
	te51_MOV_Rn_direct      when x"AE", -- MOV   R6, data addr
	te51_MOV_Rn_direct      when x"AF", -- MOV   R7, data addr
	te51_ANL_C_bit          when x"B0", -- ANL   C, bit addr
	te51_ACALL_addr11       when x"B1", -- ACALL code addr
	te51_CPL_bit            when x"B2", -- CPL   bit addr
	te51_CPL_C              when x"B3", -- CPL   C
	te51_CJNE_A_xdata_rel   when x"B4", -- CJNE  A, #data, code addr
	te51_CJNE_A_direct_rel  when x"B5", -- CJNE  A, data addr, code addr
	te51_CJNE_aRi_xdata_rel when x"B6", -- CJNE  @R0, #data, code addr
	te51_CJNE_aRi_xdata_rel when x"B7", -- CJNE  @R1, #data, code addr
	te51_CJNE_Rn_xdata_rel  when x"B8", -- CJNE  R0, #data, code addr
	te51_CJNE_Rn_xdata_rel  when x"B9", -- CJNE  R1, #data, code addr
	te51_CJNE_Rn_xdata_rel  when x"BA", -- CJNE  R2, #data, code addr
	te51_CJNE_Rn_xdata_rel  when x"BB", -- CJNE  R3, #data, code addr
	te51_CJNE_Rn_xdata_rel  when x"BC", -- CJNE  R4, #data, code addr
	te51_CJNE_Rn_xdata_rel  when x"BD", -- CJNE  R5, #data, code addr
	te51_CJNE_Rn_xdata_rel  when x"BE", -- CJNE  R6, #data, code addr
	te51_CJNE_Rn_xdata_rel  when x"BF", -- CJNE  R7, #data, code addr
	te51_PUSH_direct        when x"C0", -- PUSH  data addr
	te51_AJMP_addr11        when x"C1", -- AJMP  code addr
	te51_CLR_bit            when x"C2", -- CLR   bit addr
	te51_CLR_C              when x"C3", -- CLR   C
	te51_SWAP_A             when x"C4", -- SWAP  A
	te51_XCH_A_direct       when x"C5", -- XCH   A, data addr
	te51_XCH_A_aRi          when x"C6", -- XCH   A, @R0
	te51_XCH_A_aRi          when x"C7", -- XCH   A, @R1
	te51_XCH_A_Rn           when x"C8", -- XCH   A, R0
	te51_XCH_A_Rn           when x"C9", -- XCH   A, R1
	te51_XCH_A_Rn           when x"CA", -- XCH   A, R2
	te51_XCH_A_Rn           when x"CB", -- XCH   A, R3
	te51_XCH_A_Rn           when x"CC", -- XCH   A, R4
	te51_XCH_A_Rn           when x"CD", -- XCH   A, R5
	te51_XCH_A_Rn           when x"CE", -- XCH   A, R6
	te51_XCH_A_Rn           when x"CF", -- XCH   A, R7
	te51_POP_direct         when x"D0", -- POP   data addr
	te51_ACALL_addr11       when x"D1", -- ACALL code addr
	te51_SETB_bit           when x"D2", -- SETB  bit addr
	te51_SETB_C             when x"D3", -- SETB  C
	te51_DA_A               when x"D4", -- DA    A
	te51_DJNZ_direct_rel    when x"D5", -- DJNZ  data addr, code addr
	te51_XCHD_A_aRi         when x"D6", -- XCHD  A, @R0
	te51_XCHD_A_aRi         when x"D7", -- XCHD  A, @R1
	te51_DJNZ_Rn_rel        when x"D8", -- DJNZ  R0, code addr
	te51_DJNZ_Rn_rel        when x"D9", -- DJNZ  R1, code addr
	te51_DJNZ_Rn_rel        when x"DA", -- DJNZ  R2, code addr
	te51_DJNZ_Rn_rel        when x"DB", -- DJNZ  R3, code addr
	te51_DJNZ_Rn_rel        when x"DC", -- DJNZ  R4, code addr
	te51_DJNZ_Rn_rel        when x"DD", -- DJNZ  R5, code addr
	te51_DJNZ_Rn_rel        when x"DE", -- DJNZ  R6, code addr
	te51_DJNZ_Rn_rel        when x"DF", -- DJNZ  R7, code addr
	te51_MOVX_A_aDPTR       when x"E0", -- MOVX  A, @DPTR
	te51_AJMP_addr11        when x"E1", -- AJMP  code addr
	te51_MOVX_A_aRi         when x"E2", -- MOVX  A, @R0
	te51_MOVX_A_aRi         when x"E3", -- MOVX  A, @R1
	te51_CLR_A              when x"E4", -- CLR   A
	te51_MOV_A_direct       when x"E5", -- MOV   A, data addr
	te51_MOV_A_aRi          when x"E6", -- MOV   A, @R0
	te51_MOV_A_aRi          when x"E7", -- MOV   A, @R1
	te51_MOV_A_Rn           when x"E8", -- MOV   A, R0
	te51_MOV_A_Rn           when x"E9", -- MOV   A, R1
	te51_MOV_A_Rn           when x"EA", -- MOV   A, R2
	te51_MOV_A_Rn           when x"EB", -- MOV   A, R3
	te51_MOV_A_Rn           when x"EC", -- MOV   A, R4
	te51_MOV_A_Rn           when x"ED", -- MOV   A, R5
	te51_MOV_A_Rn           when x"EE", -- MOV   A, R6
	te51_MOV_A_Rn           when x"EF", -- MOV   A, R7
	te51_MOVX_aDPTR_A       when x"F0", -- MOVX  @DPTR, A
	te51_ACALL_addr11       when x"F1", -- ACALL code addr
	te51_MOVX_aRi_A         when x"F2", -- MOVX  @R0, A
	te51_MOVX_aRi_A         when x"F3", -- MOVX  @R1, A
	te51_CPL_A              when x"F4", -- CPL   A
	te51_MOV_direct_A       when x"F5", -- MOV   data addr, A
	te51_MOV_aRi_A          when x"F6", -- MOV   @R0, A
	te51_MOV_aRi_A          when x"F7", -- MOV   @R1, A
	te51_MOV_Rn_A           when x"F8", -- MOV   R0, A
	te51_MOV_Rn_A           when x"F9", -- MOV   R1, A
	te51_MOV_Rn_A           when x"FA", -- MOV   R2, A
	te51_MOV_Rn_A           when x"FB", -- MOV   R3, A
	te51_MOV_Rn_A           when x"FC", -- MOV   R4, A
	te51_MOV_Rn_A           when x"FD", -- MOV   R5, A
	te51_MOV_Rn_A           when x"FE", -- MOV   R6, A
	te51_MOV_Rn_A           when x"FF", -- MOV   R7, A
	te51_NOP                when others;
	
end bhv;

library IEEE;
use IEEE.STD_LOGIC_1164.all;

use work.te51pak.all;

library IEEE;
use IEEE.STD_LOGIC_1164.all;

entity te51mcode is
	port(
		clk: in STD_LOGIC;
		rst: in STD_LOGIC;
		--
		state : in te51stateT;
		instr : in te51instrT;
		--
		mcode:  out te51mcodeT
		);
end te51mcode;

--------------------------------------------------------------------------------
architecture BHV of te51mcode is
	function defaults(bytes: integer; cycles: integer) return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode.bytes := bytes;
		mcode.cycles:= cycles;
		mcode.movxrd:= '0';
		mcode.movxwr:= '0';
		
		mcode.op2   := te51op_none;
		mcode.adx21 := te51addr_none;
		mcode.adx22 := te51addr_none;
		mcode.rslt2 := '0';
		mcode.psw2  := '0';
		
		mcode.op3   := te51op_none;
		mcode.adx31 := te51addr_none;
		mcode.adx32 := te51addr_none;
		mcode.rslt3 := '0';
		mcode.psw3  := '0';
		
		return mcode;
	end function;
	
	----------------------------------------------------------------------------
	-- Mnemonic:    ACALL addr11 (p58)
	-- Function:    Absolute Call
	-- Description: ACALL unconditionally calls a subroutine located at the
	--              indicated address. The instruction increments the PC twice
	--              to obtain the address of the following instruction, then
	--              pushes the 16-bit result onto the stack (low-order byte
	--              first) and increments the Stack Pointer twice. The
	--              destination address is obtained by successively
	--              concatenating the five high-order bits of the incremented
	--              PC, obcode bits 7-5, and the second byte of the insruction.
	--              The subroutine called must therefore start within the same
	--              2K block of the program memory as the first byte of the
	--              instruction following ACALL. No flags are affected.
	-- Bytes:       2
	-- Cycles:      2
	-- Encoding:    a10 a9 a8 1 0001 | a7 a6 a5 a4 a3 a2 a1 a0
	-- Operation:   ACALL
	--              (PC)       <= (PC) + 2
	--              (SP)       <= (SP) + 1
	--              ((SP))     <= (PC(7:0))
	--              (SP)       <= (SP) + 1
	--              ((SP))     <= (PC(15:8))
	--              (PC(10:0)) <= page address
	function ACALL_addr11 return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode      := defaults(2, 2);
		mcode.op2  := te51op_add;
		mcode.adx21:= te51addr_SP;
		mcode.adx22:= te51addr_const2;
		mcode.rslt2:= '1';
		
		mcode.op3  := te51op_none;
		mcode.adx31:= te51addr_call;
		mcode.adx32:= te51addr_PC_addr11;
		mcode.rslt3:= '1';
		return mcode;
	end function;
	----------------------------------------------------------------------------
	-- Mnemonic:    ADD A, Rn (p59)
	-- Function:    Add
	-- Description: ADD adds the byte variable indicated to the Accumulator,
	--              leaving the result in the Accumulator. The carry and
	--              auxiliary-carry flags are set, respectively, if there is a
	--              carry-out from bit 7 or bit 3, and cleared otherwise. When
	--              adding unsigned integers, the carry flag indicates an
	--              overflow occured.
	--              OV is set if there is a carry-out of bit 6 but not out of
	--              bit 7, or a carry-out of bit 7 but not bit 6; otherwise
	--              OV is cleared. When adding signed integers, OV indicates a
	--              negative number produced as the sum of two positive
	--              operands, or a positive sum from two negative operands.
	--              Four source operand addressing modes are allowed: register,
	--              direct, register-indirect, or immediate.
	-- Bytes:       1
	-- Cycles:      1
	-- Encoding:    0010 1 rrr
	-- Operation:   ADD
	--              (A) <= (A) + (Rn)
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-- Mnemonic:    ADD A, direct (p59)
	-- Function:    Add
	-- Description: see above
	-- Bytes:       2
	-- Cycles:      1
	-- Encoding:    0010 0101 | direct address
	-- Operation:   ADD
	--              (A) <= (A) + (direct)
	function ADD_A_direct return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode      := defaults(2, 1);
		mcode.op2  := te51op_add;
		mcode.adx21:= te51addr_A;
		mcode.adx22:= te51addr_direct;
		mcode.rslt2:= '1';
		mcode.psw2 := '1';
		return mcode;
	end function;
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-- Mnemonic:    ADD A, @Ri (p60)
	-- Function:    Add
	-- Description: see above
	-- Bytes:       1
	-- Cycles:      1
	-- Encoding:    0010 011 i
	-- Operation:   ADD
	--              (A) <= (A) + ((Ri))
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-- Mnemonic:    ADD A, #data (p60)
	-- Function:    Add
	-- Description: see above
	-- Bytes:       2
	-- Cycles:      1
	-- Encoding:    0010 0100 | immediate data
	-- Operation:   ADD
	--              (A) <= (A) + #data
	----------------------------------------------------------------------------
	-- Mnemonic:    ADDC A, Rn
	-- Mnemonic:    ADDC A, @Ri
	-- Mnemonic:    ADDC A, #data
	----------------------------------------------------------------------------
	-- Mnemonic:    AJMP addr11 (p62)
	-- Function:    Absolute Jump
	-- Description: AJMP transfers program execution to the indicated address,
	--              which is formed at run-time by concatenating the high-order
	--              five bits of the PC (after incrementing the PC twice),
	--              opcode bits 7-5, and the second byte of the instruction.
	--              The destination must therefore be within the same 2k block
	--              of program memory as the first byte of the instruction 
	--              following AJMP.
	-- Bytes:       2
	-- Cycles:      2
	-- Encoding:    a10 a9 a8 0 0001 | a7 a6 a5 a4 a3 a2 a1 a0
	-- Operation:   AJMP
	--              (PC) <= (PC) + 2
	--              (PC10:0) <= page address
	function AJMP_addr11 return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode      := defaults(2, 2);
		mcode.op3  := te51op_none;
		mcode.adx31:= te51addr_jmp;
		mcode.adx32:= te51addr_PC_addr11;
		mcode.rslt3:= '1';
		return mcode;
	end function;
	----------------------------------------------------------------------------
	-- Mnemonic:    ANL A, Rn
	-- Function:    Logical-AND for byte variables
	-- Description: ANL performs the bitwise logial-AND operation between the
	--              variables indicated and stores the result in the destination
	--              variable. No flags are affected.
	--              The two operands allow six addressing mode combinations.
	--              When the destination is the Accumulator, the source can use
	--              register, direct, register-indirect, or immediate addressing;
	--              when the destination is a direct address, the source can be
	--              the Accumulator or immediate data.
	--              Note: When this instruction is used to modify an output port,
	--              the value used as the original port data will be read
	--              from the output data latch, not the input pins.
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-- Mnemonic:    ANL A, direct
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-- Mnemonic:    ANL A, @Ri
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-- Mnemonic:    ANL A, #data (p63)
	-- Function:    Logical-AND for byte variables
	-- Description: see above
	-- Bytes:       2
	-- Cycles:      1
	-- Encoding:    0101 0100 | immediate data
	-- Operation:   ANL
	--              (A) <= (A) and #data
	function ANL_A_xdata return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode      := defaults(2, 1);
		mcode.op2  := te51op_and;
		mcode.adx21:= te51addr_A;
		mcode.adx22:= te51addr_xdata2;
		mcode.rslt2:= '1';
		return mcode;
	end function;
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-- Mnemonic:    ANL direct, A
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-- Mnemonic:    ANL direct, #data
	----------------------------------------------------------------------------
	-- Mnemonic:    ANL C, bit (p64)
	-- Function:    Logical-AND for bit variables
	-- Description: If the Boolean value of the source bit is a logical 0 then
	--              clear the carry flag; otherwise leave the carry flag in its
	--              current state. A slash ("/") preceding the operand in the
	--              assembly language indicates that the logical complement of
	--              the addressed bit is used as the source value, but the source
	--              bit itself is not affected. No other flags are affected.
	--              Only direct addressing is allowed for the source operand.
	-- Bytes:       2
	-- Cycles:      2
	-- Encoding:    1000 0010 | bit address
	-- Operation:   ANL
	--              (C) <= (C) and (bit)
	function ANL_C_bit return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode      := defaults(2, 2);
		mcode.op2  := te51op_and;
		mcode.adx21:= te51addr_CY;
		mcode.adx22:= te51addr_bit;
		mcode.rslt2:= '1';
		return mcode;
	end function;
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-- Mnemonic:    ANL C, /bit (p64)
	-- Bytes:       2
	-- Cycles:      2
	-- Encoding:    1011 0000 | bit address
	-- Operation:   ANL
	--              (C) <= (C) and not (bit)
	----------------------------------------------------------------------------
	-- Mnemonic:    CJNE A, direct, rel
	-- Function:    Compare and Jump if Not Equal
	-- Description: CJNE compares the magnitues of the first two operands, and
	--              branches if their values are not equal. The branch
	--              destination is computed by adding the signed relative-
	--              displacement in the last instruction byte to the PC, after
	--              incrementing the PC to the start of the next instruction.
	--              The carry flag is set if the unsigned integer value of
	--              <dest-byte> is less than the unsigned integer value of
	--              <src-byte>; otherwise, the carry is cleared. Neither operand
	--              is affected.
	--              The first two operands allow four addressing mode combinations:
	--              the Accumulator may be compared with any directly addressed
	--              byte or immediate data, and any indirect RAM location or
	--              working register can be compared with an immediate constant.
	--             
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-- Mnemonic:    CJNE A, #data, rel (p66)
	-- Function:    Compare and Jump if Not Equal
	-- Description: see above
	-- Bytes:       3
	-- Cycles:      2
	-- Encoding:    1011 0100 | immediate data | rel. address
	-- Operation:   (PC) <= (PC) + 3
	--              IF (A) <> data
	--                THEN
	--                  (PC) <= (PC) + relative offset
	--              IF (A) < data
	--                THEN
	--                  (C) <= 1
	--                ELSE
	--                  (C) <= 0
	function CJNE_A_xdata_rel return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode      := defaults(3, 2);
		
		mcode.op2  := te51op_cmp;
		mcode.adx21:= te51addr_A;
		mcode.adx22:= te51addr_xdata2;
		
		mcode.op3  := te51op_clt;
		mcode.adx31:= te51addr_ne;
		mcode.adx32:= te51addr_PC_rel3;
		mcode.rslt3:= '1';
		mcode.psw3 := '1';
		return mcode;
	end function;
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-- Mnemonic:    CJNE Rn, #data, rel
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-- Mnemonic:    CJNE @Ri, #data, rel
	----------------------------------------------------------------------------
	-- Mnemonic:    CLR A (p67)
	-- Function:    Clear Accumulator
	-- Description: The Accumulator is cleared (all bits set on zero).
	--              No flags are affected.
	-- Bytes:       1
	-- Cycles:      1
	-- Encoding:    1110 0100
	-- Operation:   CLR
	--              (A) <= 0
	----------------------------------------------------------------------------
	-- Mnemonic:    CLR C (p67)
	-- Function:    Clear bit
	-- Description: The indicated bit is cleared (reset to zero). No other flags
	--              are affected. CLR can operate on the carry flag or any
	--              directly addressable bit.
	-- Bytes:       1
	-- Cycles:      1
	-- Encoding:    1100 0011
	-- Operation:   CLR
	--              (C) <= 0
	function CLR_C return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode      := defaults(1, 1);
		mcode.op2  := te51op_src2;
		mcode.adx21:= te51addr_CY;
		mcode.adx22:= te51addr_c0;
		mcode.rslt2:= '1';
		return mcode;
	end function;
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-- Mnemonic:    CLR bit (p67)
	-- Bytes:       2
	-- Cycles:      1
	-- Encoding:    1100 0010 | bit address
	-- Operation:   CLR
	--              (bit) <= 0
	----------------------------------------------------------------------------
	-- Mnemonic:    CPL A
	-- Mnemonic:    CPL C
	-- Mnemonic:    CPL bit
	-- Mnemonic:    DA A
	----------------------------------------------------------------------------
	-- Mnemonic:    DEC A
	-- Function:    Decrement
	-- Description: The variable indicated is decremented by 1. An original
	--              An original value of 00H will underflow to 0FFH. No flags
	--              are affected. Four operand addressing modes are allowed:
	--              accumulator, register, direct, or register-indirect.
	--              Note: When this instruction is used to modify an output
	--              port, the value used as the original port data will be read
	--              from the output data latch, not the input pins.
	-- Mnemonic:    DEC Rn
	-- Mnemonic:    DEC direct
	-- Mnemonic:    DEC @Ri
	function DEC_A return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode      := defaults(1, 1);
		mcode.op2  := te51op_sub;
		mcode.adx21:= te51addr_A;
		mcode.adx22:= te51addr_c1;
		mcode.rslt2:= '1';
		return mcode;
	end function;
	----------------------------------------------------------------------------
	-- Mnemonic:    DIV AB
	-- Mnemonic:    DJNZ Rn, rel
	-- Mnemonic:    DJNZ direct, rel
	----------------------------------------------------------------------------
	-- Mnemonic:    INC A
	-- Function:    Increment
	-- Description: INC increments the indicated value by 1. An original value
	--              of 0FFH will overflow to 00H. No flags are affected. Three
	--              addressing modes are allowed: register, direct or
	--              register-indirect.
	--              Note: When this instruction is used to modify an output
	--              port, the value used as the original port data will be read
	--              from the output data latch, not the input pins.
	-- Bytes:       1
	-- Cycles:      1
	-- Encoding:    0000 0100
	-- Operation:   INC
	--              (A) <= (A) + 1
	function INC_A return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode      := defaults(1, 1);
		mcode.op2  := te51op_add;
		mcode.adx21:= te51addr_A;
		mcode.adx22:= te51addr_c1;
		mcode.rslt2:= '1';
		return mcode;
	end function;
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-- Mnemonic:    INC Rn (p75)
	-- Bytes:       1
	-- Cycles:      1
	-- Encoding:    0000 1rrr
	-- Operation:   INC
	--              (Rn) <= (Rn) + 1
	function INC_Rn return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode      := defaults(1, 1);
		mcode.op2  := te51op_add;
		mcode.adx21:= te51addr_Rn;
		mcode.adx22:= te51addr_c1;
		mcode.rslt2:= '1';
		return mcode;
	end function;
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-- Mnemonic:    INC direct (p75)
	-- Bytes:       2
	-- Cycles:      1
	-- Endcoding:   0000 0101 | direct address
	-- Operation:   INC
	--              (direct) <= (direct) + 1
	function INC_direct return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode      := defaults(2, 1);
		mcode.op2  := te51op_add;
		mcode.adx21:= te51addr_direct;
		mcode.adx22:= te51addr_c1;
		mcode.rslt2:= '1';
		return mcode;
	end function;
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-- Mnemonic:    INC @Ri
	-- Bytes:       1
	-- Cycles:      1
	-- Encoding:    0000 011i
	-- Operation:   INC
	--              ((Ri)) <= ((Ri)) + 1
	function INC_aRi return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode      := defaults(1, 1);
		mcode.op2  := te51op_add;
		mcode.adx21:= te51addr_aRi;
		mcode.adx22:= te51addr_c1;
		mcode.rslt2:= '1';
		return mcode;
	end function;
	----------------------------------------------------------------------------
	-- Mnemonic:    INC DPTR (p75)
	-- Function:    Increment Data Pointer
	-- Description: Increment the 16-bit data pointer by 1. A 16-bit increment
	--              (modulo 2^16) is performed; an overflow of the low-order
	--              byte of the data pointer (DPL) from 0FFH to 00H will
	--              increment the high-order byte (DPH). No flags are affected.
	--              This is the only 16-bit register which can be incremented.
	-- Bytes:       1
	-- Cycles:      2
	-- Encoding:    1010 0011
	-- Operation:   INC
	--              (DPTR) <= (DPTR) + 1
	function INC_DPTR return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode      := defaults(1, 2);
		mcode.op2  := te51op_add;
		mcode.adx21:= te51addr_DPL;
		mcode.adx22:= te51addr_c1;
		mcode.rslt2:= '1';
		mcode.op3  := te51op_add;
		mcode.adx31:= te51addr_DPH;
		mcode.adx32:= te51addr_cy16;
		mcode.rslt3:= '1';
		return mcode;
	end function;
	----------------------------------------------------------------------------
	-- Mnemonic:    JB bit, rel
	----------------------------------------------------------------------------
	-- Mnemonic:    JC rel (p77)
	-- Function:    Jump if Carry is set
	-- Description: If the carry flag is set, branch to the address indicated;
	--              otherwise proceed with the next instruction. The branch
	--              destination is computed by adding the signed relative-
	--              displacement in the second instruction byte to the PC, after
	--              incrementing the PC twice. No flags are affected.
	-- Bytes:       2
	-- Cycles:      2
	-- Encoding:    0100 0000 | rel. address
	-- Operation:   JC
	--              (PC) <= (PC) + 2
	--              IF (C) = 1
	--                THEN
	--                  (PC) <= (PC) + rel
	function JC_rel return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode      := defaults(2, 2);
		
		mcode.op2  := te51op_none;
		mcode.adx21:= te51addr_none;
		mcode.adx22:= te51addr_none;
		
		mcode.op3  := te51op_none;
		mcode.adx31:= te51addr_jc;
		mcode.adx32:= te51addr_PC_rel2;
		mcode.rslt3:= '1';
		return mcode;
	end function;
	----------------------------------------------------------------------------
	-- Mnemonic:    JMP @A+DPTR
	----------------------------------------------------------------------------
	-- Mnemonic:    JNB bit, rel
	-- Function:    Jump if Bit Not set
	-- Description: If the indicated bit is a zero, branch to the indicated
	--              address; otherwise proceed with the next instruction.
	--              The branch destination is computed by adding the signed
	--              relative-displacement in the third instruction byte to
	--              the PC, after incrementing the PC to the first byte of
	--              the next instruction. The bit tested is not modified.
	--              No flags are affected.
	-- Bytes:       3
	-- Cycles:      2
	-- Encoding:    0011 0000 | bit address | re. address
	-- Operation:   JNB
	--              (PC) <= (PC) + 3
	--              IF (bit) = 0
	--                THEN (PC) <= (PC) + rel.
	function JNB_bit_rel return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode      := defaults(3, 2);
		
		mcode.op2  := te51op_cmp;
		mcode.adx21:= te51addr_bit;
		mcode.adx22:= te51addr_c0;
		
		mcode.op3  := te51op_none;
		mcode.adx31:= te51addr_jeq;
		mcode.adx32:= te51addr_PC_rel3;
		mcode.rslt3:= '1';
		return mcode;
	end function;
	----------------------------------------------------------------------------
	-- Mnemonic:    JNC rel (p79)
	-- Function:    Jump if Carry not set.
	-- Description: If the carry flag is a zero, branch to the address
	--              indicated; otherwise proceed with the next instruction.
	--              The branch destination is computed by adding the signed
	--              relative-displacement in the second instruction byte to
	--              the PC, after incrementing the PC twice to point to the
	--              next instruction. The carry flag is not modified.
	-- Bytes:       2
	-- Cycles:      2
	-- Encoding:    0101 0000 | rel. address
	-- Operation:   JNC
	--              (PC) <= PC + 2
	--              IF (C) = 0
	--                THEN
	--                  (PC) <= (PC) + rel
	function JNC_rel return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode      := defaults(2, 2);
		
		mcode.op2  := te51op_none;
		mcode.adx21:= te51addr_none;
		mcode.adx22:= te51addr_none;
		
		mcode.op3  := te51op_none;
		mcode.adx31:= te51addr_jnc;
		mcode.adx32:= te51addr_PC_rel2;
		mcode.rslt3:= '1';
		return mcode;
	end function;
	----------------------------------------------------------------------------
	-- Mnemonic:    JNZ rel (p80)
	-- Function:    Jump if Accumulator Not Zero
	-- Description: If any bit of the Accumulator is a one, branch to the
	--              indicated address; otherwise proceed with the next
	--              instruction. The branch destination is computed by adding
	--              the signed relative-displacement in the second instruction
	--              byte to the PC, after incrementing the PC twice. The
	--              Accumulator is not modified. No flags are affected.
	-- Bytes:       2
	-- Cycles:      2
	-- Encoding:    0111 000 | re. address
	-- Operation:   JNZ
	--              (PC) <= (PC) + 2
	--              IF (A) /= 0
	--                THEN (PC) <= (PC) + rel
	function JNZ_rel return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode      := defaults(2, 2);
		
		mcode.op2  := te51op_cmp;
		mcode.adx21:= te51addr_A;
		mcode.adx22:= te51addr_c0;
		
		mcode.op3  := te51op_none;
		mcode.adx31:= te51addr_ne;
		mcode.adx32:= te51addr_PC_rel2;
		mcode.rslt3:= '1';
		return mcode;
	end function;
	----------------------------------------------------------------------------
	-- Mnemonic:    JZ rel (p80)
	-- Function:    Jump if Accumulator Zero
	-- Description: If all bits of the Accumulator are zero, branch to the
	--              address indicated; otherwise proceed with the next
	--              instruction. The branch destination is computed by adding
	--              the signed relative-displaycement in the second instruction
	--              byte to the PC, after incrementing the PC twice. The 
	--              Accumulator is not modified. No flags are affected.
	-- Bytes:       2
	-- Cycles:      2
	-- Encoding:    0110 0000 | rel. address
	-- Operation:   JZ
	--              (PC) <= (PC) + 2
	--              IF (A) = 0
	--                THEN (PC) <= (PC) + rel
	function JZ_rel return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode      := defaults(2, 2);
		
		mcode.op2  := te51op_cmp;
		mcode.adx21:= te51addr_A;
		mcode.adx22:= te51addr_c0;
		
		mcode.op3  := te51op_none;
		mcode.adx31:= te51addr_jeq;
		mcode.adx32:= te51addr_PC_rel2;
		mcode.rslt3:= '1';
		return mcode;
	end function;
	----------------------------------------------------------------------------
	-- Mnemonic:    LCALL addr16 (p81)
	-- Function:    Long call
	-- Description: LCALL calls a subroutine located at the indicated address.
	--              The instruction adds three to the program counter to
	--              generate the address of the next instruction and then
	--              pushes the 16-bit result onto the stack (low byte first),
	--              incrementing the Stack Pointer by two. The high-order and
	--              low-order bytes of the PC are then loaded, respectively,
	--              with the second and third bytes of the LCALL instruction.
	--              Program execution continues with the instruction at this
	--              address. The subroutine may therefore begin anywhere in
	--              the full 64k-byte program memory address space. No flags
	--              are affected.
	-- Bytes:       3
	-- Cycles:      2
	-- Encoding:    0001 0010 | addr15:8 | addr7:0
	-- Operation:   LCALL
	--              (PC) <= (PC) + 3
	--              (SP) <= (SP) + 1
	--              ((SP)) <= (PC7:0)
	--              (SP) <= (SP) + 1
	--              ((SP)) <= (PC15:8)
	--              (PC) <= addr15:0
	function LCALL_addr16 return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode      := defaults(3, 2);
		mcode.op2  := te51op_add;
		mcode.adx21:= te51addr_SP;
		mcode.adx22:= te51addr_const2;
		mcode.rslt2:= '1';
		
		mcode.op3  := te51op_none;
		mcode.adx31:= te51addr_call;
		mcode.adx32:= te51addr_PC_addr16;
		mcode.rslt3:= '1';
		return mcode;
	end function;
	----------------------------------------------------------------------------
	-- Mnemonic:    LJMP addr16 (p81)
	-- Function:    Long Jump
	-- Description: LJMP causes an unconditional branch to the indicated
	--              address, by loading the high-order and low-order bytes
	--              of the PC (respectively) with the second and third
	--              instruction bytes. The destination may therefore be
	--              anywhere in the full 64k program memory address space.
	--              No flags are affected.
	-- Bytes:       3
	-- Cycles:      2
	-- Encoding:    0000 0010 | addr15-8 | addr7-0
	-- Operation:   LJMP
	--              (PC) <= addr15:0
	function LJMP_addr16 return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode      := defaults(3, 2);
		mcode.op3  := te51op_none;
		mcode.adx31:= te51addr_jmp;
		mcode.adx32:= te51addr_PC_addr16;
		mcode.rslt3:= '1';
		return mcode;
	end function;
	----------------------------------------------------------------------------
	-- Mnemonic:    MOV A, Rn
	-- Function:    Move byte variable
	-- Description: The byte variable indicated by the second operand is copied
	--              into the location specified by the first operand. The source
	--              byte is not affected. No other register or flag is affected.
	--              This is by far the most flexible operation. Fifteen
	--              combinations of source and destination addressing modes
	--              are allowed.
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-- Mnemonic:    MOV A, direct (p82)
	-- Function:    Move byte variable
	-- Description: see above
	-- Bytes:       2
	-- Cycles:      1
	-- Encoding:    1110 0101 | direct address
	-- Operation:   MOV
	--              (A) <= (direct)
	-- MOV A, ACC is not a valid instruction.
	function MOV_A_direct return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode      := defaults(2, 1);
		mcode.op2  := te51op_src2;
		mcode.adx21:= te51addr_A;
		mcode.adx22:= te51addr_direct;
		mcode.rslt2:= '1';
		return mcode;
	end function;
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-- Mnemonic:    MOV A, @Ri
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-- Mnemonic:    MOV A, #data (p83)
	-- Function:    Move byte variable
	-- Description: see above
	-- Bytes:       2
	-- Cycles:      1
	-- Encoding:    0111 0100 | immediate data
	-- Operation:   MOV
	--              (A) <= #data
	function MOV_A_xdata return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode      := defaults(2, 1);
		mcode.op2  := te51op_src2;
		mcode.adx21:= te51addr_A;
		mcode.adx22:= te51addr_xdata2;
		mcode.rslt2:= '1';
		return mcode;
	end function;
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-- Mnemonic:    MOV Rn, A
	-- Mnemonic:    MOV Rn, direct
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-- Mnemonic:    MOV Rn, #data (p83)
	-- Function:    Move byte variable
	-- Description: see above
	-- Bytes:       2
	-- Cycles:      1
	-- Encoding:    0111 1rrr | immediate data
	-- Operation    MOV
	--              (Rn) <= #data
	function MOV_Rn_xdata return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode      := defaults(2, 1);
		mcode.op2  := te51op_src2;
		mcode.adx21:= te51addr_Rn;
		mcode.adx22:= te51addr_xdata2;
		mcode.rslt2:= '1';
		return mcode;
	end function;
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-- Mnemonic:    MOV direct, A (p84)
	-- Function:    Move byte variable
	-- Description: see above
	-- Bytes:       2
	-- Cycles:      1
	-- Encoding:    1111 0101 | direct address
	-- Operation:   MOV
	--              (direct) <= (A)
	function MOV_direct_A return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode      := defaults(2, 1);
		mcode.op2  := te51op_src2;
		mcode.adx21:= te51addr_direct;
		mcode.adx22:= te51addr_A;
		mcode.rslt2:= '1';
		return mcode;
	end function;
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-- Mnemonic:    MOV direct, Rn
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-- Mnemonic:    MOV direct, direct
	-- Bytes:       3
	-- Cycles:      2
	-- Encoding:    1000 0101 | dir. addr. (src) | dir. addr (dest)
	-- Operation:   MOV
	--              (direct) <= (direct)
	function MOV_direct_direct return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode      := defaults(3, 2);
		mcode.op3  := te51op_src2;
		mcode.adx31:= te51addr_direct3;
		mcode.adx32:= te51addr_direct;
		mcode.rslt3:= '1';
		return mcode;
	end function;
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-- Mnemonic:    MOV direct, @Ri
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-- Mnemonic:    MOV direct, #data (p84)
	-- Function:    Move
	-- Description: see above
	-- Bytes:       3
	-- Cycles:      2
	-- Encoding:    0 1 1 1 0 1 0 1 | direct address | immediate data
	-- Operation:   MOV
	--              (direct) <= #data
	function MOV_direct_xdata return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode      := defaults(3, 2);
		mcode.op3  := te51op_src2;
		mcode.adx31:= te51addr_direct;
		mcode.adx32:= te51addr_xdata3;
		mcode.rslt3:= '1';
		return mcode;
	end function;
	----------------------------------------------------------------------------
	-- Mnemonic:    MOV @Ri, A
	-- Mnemonic:    MOV @Ri, direct
	-- Mnemonic:    MOV @Ri, #data
	----------------------------------------------------------------------------
	-- Mnemonic:    MOV C, bit (p86)
	-- Function:    Move bit data
	-- Description: The Boolean variable indicated by the second operand is
	--              copied into the location specified by the first operand.
	--              One of the operands must be the carry flag; the other may
	--              be any directly addressable bit. No other register or flag
	--              is affected.
	-- Bytes:       2
	-- Cycles:      1
	-- Encoding:    1010 0010 | bit address
	-- Operation:   MOV
	--              (C) <= (bit)
	function MOV_C_bit return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode      := defaults(2, 1);
		mcode.op2  := te51op_src2;
		mcode.adx21:= te51addr_CY;
		mcode.adx22:= te51addr_bit;
		mcode.rslt2:= '1';
		return mcode;
	end function;
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-- Mnemonic:    MOV bit, C (p86)
	-- Bytes:       2
	-- Cycles:      2  <<< is this a typo???
	-- Encoding:    1001 0010 | bit address
	-- Operation:   MOV
	--              (bit) <= (C)
	function MOV_bit_C return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode      := defaults(2, 2);
		mcode.op2  := te51op_src2;
		mcode.adx21:= te51addr_bit;
		mcode.adx22:= te51addr_CY;
		mcode.rslt2:= '1';
		return mcode;
	end function;
	----------------------------------------------------------------------------
	-- Mnemonic:    MOV DPTR, #data16 (p86)
	-- Function:    Move
	-- Description: Load Data Pointer with a 16-bit constant.
	-- Bytes:       3
	-- Cycles:      2
	-- Encoding:    1001 0000 | immed. data 15-8 | immed. data 7-0
	function MOV_DPTR_xdata16 return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode      := defaults(3, 2);
		
		mcode.op2  := te51op_src2;
		mcode.adx21:= te51addr_DPH;
		mcode.adx22:= te51addr_xdata2;
		mcode.rslt2:= '1';
		
		mcode.op3  := te51op_src2;
		mcode.adx31:= te51addr_DPL;
		mcode.adx32:= te51addr_xdata3;
		mcode.rslt3:= '1';
		return mcode;
	end function;
	----------------------------------------------------------------------------
	-- Mnemonic:    MOV A, @A+DPTR
	-- Mnemonic:    MOV A, @A+PC
	----------------------------------------------------------------------------
	-- Mnemonic:    MOVX A, @Ri (p89)
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-- Mnemonic:    MOVX A, @DPTR (p89)
	-- Function:    Move External
	-- Description: see above
	-- Bytes:       1
	-- Cycles:      2
	-- Encoding:    1110 0000
	-- Operation:   MOVX
	--              (A) <= (DPTR)
	function MOVX_A_aDPTR return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode       := defaults(1, 2);
		mcode.movxrd:= '1';
		mcode.adx31 := te51addr_datain;
		mcode.rslt3 := '1';
		return mcode;
	end function;
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-- Mnemonic:    MOVX @Ri, A (p89)
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-- Mnemonic:    MOVX @DPTR, A (p89)
	-- Function:    Move External
	-- Description: see above
	-- Bytes:       1
	-- Cycles:      2
	-- Encoding:    1111 0000
	-- Operation:   MOVX
	--              (DPTR) <= (A)
	function MOVX_aDPTR_A return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode       := defaults(1, 2);
		mcode.movxwr:= '1';
		return mcode;
	end function;
	----------------------------------------------------------------------------
	-- Mnemonic:    MUL AB
	----------------------------------------------------------------------------
	-- Mnemonic:    NOP (p90)
	-- Function:    No Operation
	-- Description: Execution continues at the following instruction. 
	--              Other than the PC, no registers or flags are affected.
	-- Bytes:       1
	-- Cycles:      1
	-- Encoding:    0 0 0 0 | 0 0 0 0
	-- Operation:   NOP
	--              (PC) <= (PC) + 1
	function NOP return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode:= defaults(1, 1);
		return mcode;
	end function;
	----------------------------------------------------------------------------
	-- Mnemonic:    ORL A, Rn
	-- Function:    Logical-OR for byte variables
	-- Description: ORL performs the bitwise logical-OR operation between the
	--              indicated variables, storing the results in the destination
	--              byte. No flags are affected.
	--              The two operands allow six addressing mode combinations.
	--              When the destination is the Accumulator, the source can use
	--              register, direct, register-indirect, or immediate
	--              addressing; when the destination is a direct address, the
	--              source can be the Accumulator or immediate data.
	--              Note: When this instruction is used to modify an output
	--              port, the value used as the original port data will be read
	--              from the output data latch, not the input pins.
	
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-- Mnemonic:    ORL A, direct (p92)
	-- Function:    Logical-OR for byte variables
	-- Description: see above
	-- Bytes:       2
	-- Cycles:      1
	-- Encoding:    0100 0101 | direct address
	-- Operation:   ORL
	--              (A) <= (A) V (direct)
	function ORL_A_direct return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode      := defaults(2, 1);
		mcode.op2  := te51op_or;
		mcode.adx21:= te51addr_A;
		mcode.adx22:= te51addr_direct;
		mcode.rslt2:= '1';
		return mcode;
	end function;
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-- Mnemonic:    ORL A, @Ri
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-- Mnemonic:    ORL A, #data (p92)
	-- Function:    Logical-OR for byte variables
	-- Description: see above
	-- Bytes:       2
	-- Cycles:      1
	-- Encoding:    0100 0100 | immediate data
	-- Operation:   ORL
	--              (A) <= (A) V #data
	function ORL_A_xdata return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode      := defaults(2, 1);
		mcode.op2  := te51op_or;
		mcode.adx21:= te51addr_A;
		mcode.adx22:= te51addr_xdata2;
		mcode.rslt2:= '1';
		return mcode;
	end function;
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-- Mnemonic:    ORL direct, A
	-- Mnemonic:    ORL direct, #data
	-- Mnemonic:    ORL C, bit
	-- Mnemonic:    ORL C, /bit
	----------------------------------------------------------------------------
	-- Mnemonic:    POP direct (p94)
	----------------------------------------------------------------------------
	-- Mnemonic:    PUSH direct (p94)
	-- Function:    Push onto stack
	-- Description: The Stack Pointer is incremented by one. The contents of
	--              the indicated variable is then copied into the internal
	--              RAM located address by the Stack Pointer. Otherwise no
	--              flags are affected.
	-- Bytes:       2
	-- Cycles:      2
	-- Encoding:    1100 0000 | direct address
	-- Operation:   PUSH
	--              (SP) <= (SP) + 1
	--              ((SP)) <= (direct)
	function PUSH_direct return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode      := defaults(2, 2);
		mcode.op2  := te51op_add;
		mcode.adx21:= te51addr_SP;
		mcode.adx22:= te51addr_c1;
		mcode.rslt2:= '1';
		
		mcode.op3  := te51op_src2;
		mcode.adx31:= te51addr_aSP;
		mcode.adx32:= te51addr_direct;
		mcode.rslt3:= '1';
		return mcode;
	end function;
	----------------------------------------------------------------------------
	-- Mnemonic:    RET (p95)
	-- Function:    Return from subroutine
	-- Description: RET pops the high- and low-order bytes of the PC
	--              successively from the stack, decrementing the Stack Pointer
	--              by two. Program execution continues at the resulting
	--              address, generally the instruction immediately following an
	--              ACALL or LCALL. No flags are affected.
	-- Bytes:       1
	-- Cycles:      2
	-- Encoding:    0010 0010
	-- Operation:   RET
	--              (PC15:8) <= ((SP))
	--              (SP) <= (SP) - 1
	--              (PC7:0) <= ((SP))
	--              (SP) <= (SP) - 1
	function RET return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode      := defaults(1, 2);
		mcode.op2  := te51op_sub;
		mcode.adx21:= te51addr_SP;
		mcode.adx22:= te51addr_const2;
		mcode.rslt2:= '1';
		
		mcode.op3  := te51op_none;
		mcode.adx31:= te51addr_ret; -- write pc_wren during alurslt
		mcode.adx32:= te51addr_ret; -- fetch pc15:8 during alusrc #2, fetch pc7:0 during aluind
		mcode.rslt3:= '1';
		return mcode;
	end function;
	----------------------------------------------------------------------------
	-- Mnemonic:    RETI
	-- Mnemonic:    RL A
	-- Mnemonic:    RLC A
	-- Mnemonic:    RR A
	-- Mnemonic:    RRC A
	----------------------------------------------------------------------------
	-- Mnemonic:    SETB C (p98)
	-- Function:    Set Bit
	-- Description: SETB sets the indicated bit to one. SETB can operate on the
	--              carry flag or any directly addressable bit. No other flags
	--              are affected.
	-- Bytes:       1
	-- Cycles:      1
	-- Encoding:    1101 0011
	-- Operation:   SETB
	--              (C) <= 1
	function SETB_C return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode      := defaults(1, 1);
		mcode.op2  := te51op_src2;
		mcode.adx21:= te51addr_CY;
		mcode.adx22:= te51addr_c1;
		mcode.rslt2:= '1';
		return mcode;
	end function;
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-- Mnemonic:    SETB bit (p98)
	-- Bytes:       2
	-- Cycles:      1
	-- Encoding:    1101 0010 | bit address
	-- Operation:   SETB
	--              (bit) <= 1
	----------------------------------------------------------------------------
	-- Mnemonic:    SJMP rel (p99)
	-- Function:    Short Jump
	-- Description: Program control branches unconditionally to the address
	--              indicated. The branch destination is computed by adding
	--              the signed displacement in the second instruction byte
	--              to the PC, after incrementing the PC twice. Therefore,
	--              the range of destinations allowed is from 128 bytes
	--              preceding this instruction to 127 bytes following it.
	-- Bytes:       2
	-- Cycles:      2
	-- Encoding:    1000 0000 | rel. address
	-- Operation:   SJMP
	--              (PC) <= (PC) + 2
	--              (PC) <= (PC) + rel
	function SJMP_rel return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode      := defaults(2, 2);
		
		mcode.op3  := te51op_none;
		mcode.adx31:= te51addr_jmp;
		mcode.adx32:= te51addr_PC_rel2;
		mcode.rslt3:= '1';
		return mcode;
	end function;
	----------------------------------------------------------------------------
	-- Mnemonic:    SUBB A, Rn (p100)
	-- Function:    Subtract with borrow
	-- Description: SUBB subtracts the indicated variable and the carry flag
	--              together from the Accumulator, leaving the result in the
	--              Accumulator. SUBB sets the carry (borrow) flag if a borrow
	--              is needed for bit 7, and clears C otherwise. (If C was set
	--              before executing a SUBB instruction, this indicates that a
	--              borrow was needed for the previous step in a multiple
	--              precision subtraction, so the carry is subtracted from the
	--              Accumulator along with the source operand.) AC is set if a
	--              borrow is needed for bit 3, and cleared otherwise. OV is
	--              set if a borrow is needed into bit 6, but not into bit 7,
	--              or into bit 7, but not bit 6.
	--              When subtracting signed integers OV indicates a negative
	--              number produces when a negative value is subtracted from
	--              a positive value, or a positive result when a positive
	--              number is subtracted from a negative number.
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-- Mnemonic:    SUBB A, direct (p101)
	-- Bytes:       2
	-- Cycles:      1
	-- Encoding:    1001 0101 | direct address
	-- Operation:   SUBB
	--              (A) <= (A) - (C) - (direct)
	function SUBB_A_direct return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode      := defaults(2, 1);
		mcode.op2  := te51op_subb;
		mcode.adx21:= te51addr_A;
		mcode.adx22:= te51addr_direct;
		mcode.rslt2:= '1';
		mcode.psw2 := '1';
		return mcode;
	end function;
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-- Mnemonic:    SUBB A, @Ri
	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	-- Mnemonic:    SUBB A, #data
	-- Bytes:       2
	-- Cycles:      1
	-- Encoding:    1001 0100 | immediate data
	-- Operation:   SUBB
	--              (A) <= (A) - (C) - #data
	function SUBB_A_xdata return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode      := defaults(2, 1);
		mcode.op2  := te51op_subb;
		mcode.adx21:= te51addr_A;
		mcode.adx22:= te51addr_xdata2;
		mcode.rslt2:= '1';
		mcode.psw2 := '1';
		return mcode;
	end function;
	----------------------------------------------------------------------------
	-- Mnemonic:    SWAP A (p101)
	-- Function:    Swap nibbles within the Accumulator
	-- Description: SWAP A interchanges the low- and high-order nibbles (four
	--              bit fields) of the Accumulator (bits 3-0 and bits 7-4). The
	--              operation can also be thought of as a four-bit rotate
	--              instruction. No flags are affected.
	-- Bytes:       1
	-- Cycles:      1
	-- Encoding:    1100 0100
	-- Operation:   SWAP
	--              (A3-0) <=> (A7-4)
	function SWAP_A return te51mcodeT is
		variable mcode: te51mcodeT;
	begin
		mcode      := defaults(1, 1);
		mcode.op2  := te51op_swap;
		mcode.adx21:= te51addr_A;
		mcode.adx22:= te51addr_none;
		mcode.rslt2:= '1';
		return mcode;
	end function;
	----------------------------------------------------------------------------
	-- Mnemonic:    XCH A, Rn
	-- Mnemonic:    XCH A, direct
	-- Mnemonic:    XCH A, @Ri
	-- Mnemonic:    XCHD A, @Ri
	-- Mnemonic:    XRL A, Rn
	-- Mnemonic:    XRL A, direct
	-- Mnemonic:    XRL A, @Ri
	-- Mnemonic:    XRL A, #data
	-- Mnemonic:    XRL direct, A
	-- Mnemonic:    XRL direct, #data
	----------------------------------------------------------------------------
begin
	
	with instr select mcode<= 
	ACALL_addr11               when te51_ACALL_addr11,
	-- ADD A, Rn (p59)
	ADD_A_direct               when te51_ADD_A_direct,
	-- ADD A, @Ri (p60)
	-- ADD A, #data (p60)
	-- ADDC A, Rn
	-- ADDC A, @Ri
	-- ADDC A, #data
	AJMP_addr11                when te51_AJMP_addr11,
	-- ANL A, Rn
	-- ANL A, direct
	-- ANL A, @Ri
	ANL_A_xdata                when te51_ANL_A_xdata,
	-- ANL direct, A
	-- ANL direct, #data
	ANL_C_bit                  when te51_ANL_C_bit,
	-- ANL C, /bit
	-- CJNE A, direct, rel
	CJNE_A_xdata_rel           when te51_CJNE_A_xdata_rel,
	-- CJNE Rn, #data, rel
	-- CJNE @Ri, #data, rel
	-- CLR A
	CLR_C                      when te51_CLR_C,
	-- CLR bit
	-- CPL A
	-- CPL C
	-- CPL bit
	-- DA A
	DEC_A                      when te51_DEC_A,
	-- DEC Rn
	-- DEC direct
	-- DEC @Ri
	-- DIV AB
	-- DJNZ Rn, rel
	-- DJNZ direct, rel
	INC_A                      when te51_INC_A,
	INC_Rn                     when te51_INC_Rn,
	INC_direct                 when te51_INC_direct,
	INC_aRi                    when te51_INC_aRi,
	INC_DPTR                   when te51_INC_DPTR,
	-- JB bit, rel
	JC_rel                     when te51_JC_rel,
	-- JMP @A+DPTR
	JNB_bit_rel                when te51_JNB_bit_rel,
	JNC_rel                    when te51_JNC_rel,
	JNZ_rel                    when te51_JNZ_rel,
	JZ_rel                     when te51_JZ_rel,
	LCALL_addr16               when te51_LCALL_addr16,
	LJMP_addr16                when te51_LJMP_addr16,
	-- MOV A, Rn
	MOV_A_direct               when te51_MOV_A_direct,
	-- MOV A, @Ri
	MOV_A_xdata                when te51_MOV_A_xdata,
	-- MOV Rn, A
	-- MOV Rn, direct
	MOV_Rn_xdata               when te51_MOV_Rn_xdata,
	MOV_direct_A               when te51_MOV_direct_A,
	-- MOV direct, Rn
	MOV_direct_direct          when te51_MOV_direct_direct,
	-- MOV direct, @Ri
	MOV_direct_xdata           when te51_MOV_direct_xdata,
	-- MOV @Ri, A
	-- MOV @Ri, direct
	-- MOV @Ri, #data
	MOV_C_bit                  when te51_MOV_C_bit,
	MOV_bit_C                  when te51_MOV_bit_C,
	MOV_DPTR_xdata16           when te51_MOV_DPTR_xdata16,
	-- MOV A, @A+DPTR
	-- MOV A, @A+PC
	-- MOVX A, @Ri
	MOVX_A_aDPTR               when te51_MOVX_A_aDPTR,
	-- MOVX @Ri, A
	MOVX_aDPTR_A               when te51_MOVX_aDPTR_A,
	-- MUL AB
	NOP                        when te51_NOP,
	-- ORL A, Rn
	ORL_A_direct               when te51_ORL_A_direct,
	-- ORL A, @Ri
	ORL_A_xdata                when te51_ORL_A_xdata,
	-- ORL direct, A
	-- ORL direct, #data
	-- ORL C, bit
	-- ORL C, /bit
	-- POP direct
	PUSH_direct                when te51_PUSH_direct,
	RET                        when te51_RET,
	-- RETI
	-- RL A
	-- RLC A
	-- RR A
	-- RRC A
	SETB_C                     when te51_SETB_C,
	-- SETB bit
	SJMP_rel                   when te51_SJMP_rel,
	-- SUBB A, Rn
	SUBB_A_direct              when te51_SUBB_A_direct,
	-- SUBB A, @Ri
	SUBB_A_xdata               when te51_SUBB_A_xdata,
	SWAP_A                     when te51_SWAP_A,
	-- XCH A, Rn
	-- XCH A, direct
	-- XCH A, @Ri
	-- XCHD A, @Ri
	-- XRL A, Rn
	-- XRL A, direct
	-- XRL A, @Ri
	-- XRL A, #data
	-- XRL direct, A
	-- XRL direct, #data
	NOP                        when others;
end BHV;

library IEEE;
use IEEE.NUMERIC_STD.all;

use work.te51pak.all;

library IEEE;
use IEEE.STD_LOGIC_1164.all;

entity te51alu is
	port(
		src1 : in STD_LOGIC_VECTOR(7 downto 0);
		src2 : in STD_LOGIC_VECTOR(7 downto 0);
		srcf : in STD_LOGIC_VECTOR(7 downto 0);
		aluop : in te51opT;
		dst1: out STD_LOGIC_VECTOR(7 downto 0);
		dstf: out STD_LOGIC_VECTOR(7 downto 0)
		);
end te51alu;

--------------------------------------------------------------------------------
architecture BHV of te51alu is
	signal src1u, src2u, srcfu: UNSIGNED(8 downto 0);
	signal srccy, srcac, srcov: STD_LOGIC;
	signal dst1u:               UNSIGNED(8 downto 0);
	signal dstcy, dstac, dstov: STD_LOGIC;
begin
	src1u<= '0' & UNSIGNED(src1);
	src2u<= '0' & UNSIGNED(src2);
	srcfu<= '0' & "0000000" & srccy;
	srccy<= srcf(7);
	srcac<= srcf(6);
	srcov<= srcf(2);
	
	process(aluop, 
		src1, src1u, 
		src2, src2u, 
		srcf, srcfu, 
		srccy, srcac, srcov, 
		dst1u)
	begin
		dstcy<= srccy;
		dstac<= srcac;
		dstov<= srcov;
		
		case aluop is
			when te51op_add=>
			dst1u<= src1u + src2u;
			dstcy<= dst1u(8);
			dstac<= '0';
			dstov<= '0';
			
			when te51op_addc=>
			dst1u<= src1u + src2u + srcfu;
			dstcy<= dst1u(8); 
			dstac<= '0';
			dstov<= '0';
			
			when te51op_sub=>
			dst1u<= src1u - src2u;
			dstcy<= dst1u(8);
			dstac<= '0';
			dstov<= '0';
			
			when te51op_subb=>
			dst1u<= src1u - src2u - srcfu;
			dstcy<= dst1u(8);
			dstac<= '0';
			dstov<= '0';
			
			when te51op_src2=>
			dst1u<= src2u;
			
			when te51op_and=>
			dst1u<= src1u and src2u;
			
			when te51op_or=>
			dst1u<= src1u or src2u;
			
			when te51op_cmp=>
			if src1=  src2 then dst1u(0)<= '1'; else dst1u(0)<= '0'; end if;
			if src1/= src2 then dst1u(1)<= '1'; else dst1u(1)<= '0'; end if;
			if src1>  src2 then dst1u(2)<= '1'; else dst1u(2)<= '0'; end if;
			if src1>= src2 then dst1u(3)<= '1'; else dst1u(3)<= '0'; end if;
			if src1<  src2 then dst1u(4)<= '1'; else dst1u(4)<= '0'; end if;
			if src1<= src2 then dst1u(5)<= '1'; else dst1u(5)<= '0'; end if;
			dst1u(7 downto 6)<= "00";
			
			when te51op_clt=>
			dstcy<= src1(4);
			
			when te51op_swap=>
			dst1u<= "0" & src1u(3 downto 0) & src1u(7 downto 4);
			
			when others=>
			dst1u<= (others=> '-');
			
		end case;
	end process;
	
	dst1<= STD_LOGIC_VECTOR(dst1u(7 downto 0));
	dstf<= dstcy & dstac & srcf(5 downto 3) & dstov & srcf(1 downto 0);
end BHV;

library IEEE;
use IEEE.NUMERIC_STD.all;

library IEEE;
use IEEE.STD_LOGIC_1164.all;

entity te51regs is
	port(
		clk:  in  STD_LOGIC;                        -- clock
		rst:  in  STD_LOGIC;                        -- reset
		adx:  in  STD_LOGIC_VECTOR(7 downto 0);     -- address
		drd:  out STD_LOGIC_VECTOR(7 downto 0);     -- data read
		dwr:  in  STD_LOGIC_VECTOR(7 downto 0);     -- data write
		badx: in  STD_LOGIC;                        -- bit address
		radx: in  STD_LOGIC;                        -- register address
		wren: in  STD_LOGIC;                        -- write enable
		--
		ao:   out STD_LOGIC_VECTOR(7 downto 0);     -- Akkumulator
		dptro:out STD_LOGIC_VECTOR(15 downto 0);    -- DPTR
		pswo: out STD_LOGIC_VECTOR(7 downto 0);     -- Program Status Word
		spo:  out STD_LOGIC_VECTOR(7 downto 0);     -- stack pointer
		--
		dbga: in  STD_LOGIC_VECTOR(7 downto 0);     -- address
		dbgv: out STD_LOGIC_VECTOR(7 downto 0)      -- value
		);
end te51regs;

--------------------------------------------------------------------------------
architecture BHV of te51regs is
	type   ramT     is array (0 to 255) of STD_LOGIC_VECTOR(7 downto 0);
	signal ram:     ramT;
	signal adxmv:   STD_LOGIC_VECTOR(adx'range);
	signal adxbit:  STD_LOGIC_VECTOR(7 downto 0);
	signal adxsfr:  STD_LOGIC_VECTOR(7 downto 0);
	signal adxreg:  STD_LOGIC_VECTOR(7 downto 0);
	signal adxrow:  UNSIGNED(7 downto 0);
	signal adxcol:  UNSIGNED(2 downto 0);
	signal drd2:    STD_LOGIC_VECTOR(7 downto 0);
	signal dwr2:    STD_LOGIC_VECTOR(7 downto 0);
	signal dwrbit:  STD_LOGIC_VECTOR(7 downto 0);
	--
	alias  a:       STD_LOGIC_VECTOR(7 downto 0) is ram(16#E0#); -- Accumulator
	alias  b:       STD_LOGIC_VECTOR(7 downto 0) is ram(16#F0#); -- B Register
	alias  psw:     STD_LOGIC_VECTOR(7 downto 0) is ram(16#D0#); -- Program Status Word
	alias  sp:      STD_LOGIC_VECTOR(7 downto 0) is ram(16#81#); -- Stack Pointer
	alias  dpl:     STD_LOGIC_VECTOR(7 downto 0) is ram(16#82#); -- Data Pointer, Low Byte
	alias  dph:     STD_LOGIC_VECTOR(7 downto 0) is ram(16#83#); --               High Byte
	alias  p0:      STD_LOGIC_VECTOR(7 downto 0) is ram(16#80#); -- Port 0
	alias  p1:      STD_LOGIC_VECTOR(7 downto 0) is ram(16#90#); -- Port 1
	alias  p2:      STD_LOGIC_VECTOR(7 downto 0) is ram(16#A0#); -- Port 2
	alias  p3:      STD_LOGIC_VECTOR(7 downto 0) is ram(16#B0#); -- Port 3
	alias  ip:      STD_LOGIC_VECTOR(7 downto 0) is ram(16#B8#); -- Interrupt Priority Control
	alias  ie:      STD_LOGIC_VECTOR(7 downto 0) is ram(16#A8#); -- Interrupt Enable Control
	alias  tmod:    STD_LOGIC_VECTOR(7 downto 0) is ram(16#89#); -- Timer/ Counter Mode Control
	alias  tcon:    STD_LOGIC_VECTOR(7 downto 0) is ram(16#88#); -- Timer/ Counter Control
	alias  tcon2:   STD_LOGIC_VECTOR(7 downto 0) is ram(16#C8#); -- Timer/ Counter 2 Control
	alias  th0:     STD_LOGIC_VECTOR(7 downto 0) is ram(16#8C#); -- Timer/ Counter 0 High Byte
	alias  tl0:     STD_LOGIC_VECTOR(7 downto 0) is ram(16#8A#); -- Timer/ Counter 0 Low Byte
	alias  th1:     STD_LOGIC_VECTOR(7 downto 0) is ram(16#8D#); -- Timer/ Counter 1 High Byte
	alias  tl1:     STD_LOGIC_VECTOR(7 downto 0) is ram(16#8B#); -- Timer/ Counter 1 Low Byte
	alias  th2:     STD_LOGIC_VECTOR(7 downto 0) is ram(16#CD#); -- Timer/ Counter 2 High Byte
	alias  tl2:     STD_LOGIC_VECTOR(7 downto 0) is ram(16#CC#); -- Timer/ Counter 2 Low Byte
	alias  rcap2h:  STD_LOGIC_VECTOR(7 downto 0) is ram(16#CB#); -- T/C 2 Capture Reg. High Byte
	alias  rcap2l:  STD_LOGIC_VECTOR(7 downto 0) is ram(16#CA#); -- T/C 2 Capture Reg. Low Byte
	alias  scon:    STD_LOGIC_VECTOR(7 downto 0) is ram(16#98#); -- Serial Control
	alias  sbuf:    STD_LOGIC_VECTOR(7 downto 0) is ram(16#99#); -- Serial Data Buffer
	alias  pcon:    STD_LOGIC_VECTOR(7 downto 0) is ram(16#87#); -- Power Control
begin
	
	------------------------------------
	-- 80h - FFh: Special Function Registers
	-- 30h - 7Fh: Scratch Pad Area
	-- 20h - 2Fh: Bit-Addressable Space (Bit Addresses 0-7F)
	-- 18h - 1Fh: register bank "11"
	-- 10h - 17h: register bank "10"
	-- 08h - 0Fh: register bank "01"
	-- 00h - 07h: register bank "00" <= reset value of stack is 07h
	
	------------------------------------
	-- strip metavalues
	adxmv<= "00000000" when adx= "UUUUUUUU" or adx= "ZZZZZZZZ" or adx= "--------"
	else    adx;
	
	------------------------------------
	-- row and column addresses
	adxbit<= "0010" & adxmv(6 downto 3);                   -- Bit-Addressable Space
	adxsfr<= "1"    & adxmv(6 downto 3) & "000";           -- bit-addressed SFRs
	adxreg<= "000"  & psw(4) & psw(3) & adxmv(2 downto 0); -- register
	
	adxrow<= UNSIGNED(adxmv)  when badx= '0' and radx= '0'
	else     UNSIGNED(adxbit) when badx= '1' and adxmv(7)= '0'
	else     UNSIGNED(adxsfr) when badx= '1' and adxmv(7)= '1'
	else     UNSIGNED(adxreg) when radx= '1';
	
	adxcol<= UNSIGNED(adxmv(2 downto 0));
	
	------------------------------------
	-- data read
	drd2<= ram(TO_INTEGER(adxrow));
	
	drd<= drd2 when badx= '0'
	--else  (0=> drd2(TO_INTEGER(adxcol)), others=> '-');
	else  (0=> drd2(TO_INTEGER(adxcol)), others=> '0');
	
	ao   <= a;
	dptro<= dph & dpl;
	pswo <= psw;
	spo  <= sp;
	
	------------------------------------
	-- data write
	--	dwrbit_gen: for i in dwr'range generate
	--		dwrbit(i)<= dwr(0) when i= adxcol else drd2(i);
	--	end generate;
	dwrbit_gen: block begin
		dwrbit(0)<= dwr(0) when adxcol= "000" else drd2(0);
		dwrbit(1)<= dwr(0) when adxcol= "001" else drd2(1);
		dwrbit(2)<= dwr(0) when adxcol= "010" else drd2(2);
		dwrbit(3)<= dwr(0) when adxcol= "011" else drd2(3);
		dwrbit(4)<= dwr(0) when adxcol= "100" else drd2(4);
		dwrbit(5)<= dwr(0) when adxcol= "101" else drd2(5);
		dwrbit(6)<= dwr(0) when adxcol= "110" else drd2(6);
		dwrbit(7)<= dwr(0) when adxcol= "111" else drd2(7);
	end block;
	
	dwr2<= dwr when badx= '0'
	else   dwrbit;
	
	------------------------------------
	-- byte-wide RAM
	process(clk, rst, adxrow)
	begin
		if rst= '1' then
			for i in 1 to 255 loop
				ram(i)<= "UUUUUUUU";
			end loop;
			
			a     <= "00000000";
			b     <= "00000000";
			psw   <= "00000000";
			sp    <= "00000111";
			dph   <= "00000000";
			dpl   <= "00000000";
			p0    <= "11111111";
			p1    <= "11111111";
			p2    <= "11111111";
			p3    <= "11111111";
			ip    <= "XXX00000";
			ie    <= "0XX00000";
			tmod  <= "00000000";
			tcon  <= "00000000";
			tcon2 <= "00000000";
			th0   <= "00000000";
			tl0   <= "00000000";
			th1   <= "00000000";
			tl1   <= "00000000";
			th2   <= "00000000";
			tl2   <= "00000000";
			rcap2h<= "00000000";
			rcap2l<= "00000000";
			scon  <= "00000000";
			sbuf  <= "XXXXXXXX";
			pcon  <= "0XXX0000";
		elsif falling_edge(clk) then
			if wren= '1' then
				ram(TO_INTEGER(adxrow))<= dwr2;
				--ram(TO_INTEGER(adxrow))<= STD_LOGIC_VECTOR(TO_UNSIGNED(TO_INTEGER(UNSIGNED(dwr2)), dwr2'length));
			end if;
		end if;
	end process;
	
	dbgv<= (others=> '0') when dbga= "ZZZZZZZZ"
	else    ram(TO_INTEGER(UNSIGNED(dbga)));
end BHV;

use work.te51pak.all;

library IEEE;
use IEEE.NUMERIC_STD.all;

library IEEE;
use IEEE.STD_LOGIC_1164.all;

entity te51ctrl is
	port(
		-- clock and reset
		clk : in STD_LOGIC;
		rst : in STD_LOGIC;
		--
		state : out te51stateT;
		-- addr/ data bus
		addr : out STD_LOGIC_VECTOR(15 downto 0);
		datai : in STD_LOGIC_VECTOR(7 downto 0);
		datao: out STD_LOGIC_VECTOR(7 downto 0);
		-- microcode
		mcode : in te51mcodeT;
		opcode1 : out STD_LOGIC_VECTOR(7 downto 0);
		-- alu
		alu_op:   out te51opT;
		alu_src1: out STD_LOGIC_VECTOR(7 downto 0);
		alu_src2: out STD_LOGIC_VECTOR(7 downto 0);
		alu_srcf: out STD_LOGIC_VECTOR(7 downto 0);
		alu_dst1: in  STD_LOGIC_VECTOR(7 downto 0);
		alu_dstf: in  STD_LOGIC_VECTOR(7 downto 0);
		-- register file
		reg_adx:  out STD_LOGIC_VECTOR(7 downto 0);
		reg_drd:  in  STD_LOGIC_VECTOR(7 downto 0);
		reg_dwr:  out STD_LOGIC_VECTOR(7 downto 0);
		reg_wren: out STD_LOGIC;
		reg_badx: out STD_LOGIC;
		reg_radx: out STD_LOGIC;
		reg_a:    in  STD_LOGIC_VECTOR(7 downto 0);
		reg_dptr: in  STD_LOGIC_VECTOR(15 downto 0);
		reg_psw:  in  STD_LOGIC_VECTOR(7 downto 0);
		reg_sp:   in  STD_LOGIC_VECTOR(7 downto 0);
		-- external memory
		ale:      out STD_LOGIC;
		psen:     out STD_LOGIC;
		wr:       out STD_LOGIC;
		rd:       out STD_LOGIC;
		-- p0 control
		p0adx:    out STD_LOGIC;
		p0oe:     out STD_LOGIC
		);
end te51ctrl;

--------------------------------------------------------------------------------
architecture bhv of te51ctrl is
	signal init:      STD_LOGIC;
	signal stateii:   te51stateT;
	signal pc:        STD_LOGIC_VECTOR(15 downto 0);
	signal pc_wren:   STD_LOGIC;
	--	signal pc_wren1, pc_wren2, pc_wren3, pc_wren4: STD_LOGIC;
	signal opcodeii:  STD_LOGIC_VECTOR(23 downto 0);
	alias  opcode1ii: STD_LOGIC_VECTOR(7 downto 0) is opcodeii(7 downto 0);
	alias  opcode2ii: STD_LOGIC_VECTOR(7 downto 0) is opcodeii(15 downto 8);
	alias  opcode3ii: STD_LOGIC_VECTOR(7 downto 0) is opcodeii(23 downto 16);
	
	signal dst1s_r, dstfs_r: STD_LOGIC_VECTOR(7  downto 0); -- result last alu slot
	signal dst1c_r, dstfc_r: STD_LOGIC_VECTOR(7  downto 0); -- result last alu clock
	signal tmp, tmp_r:       STD_LOGIC_VECTOR(15 downto 0); -- temp reg for complex addressing
	signal src2, src2_r:     STD_LOGIC_VECTOR(7  downto 0); -- source #2 mux & register
	
	constant constant_2: STD_LOGIC_VECTOR(7 downto 0):= x"02";
	constant constant_1: STD_LOGIC_VECTOR(7 downto 0):= x"01";
	constant constant_0: STD_LOGIC_VECTOR(7 downto 0):= x"00";
begin
	
	------------------------------------
	-- Timing (A) without a MOVX
	-- Cycle  |                       1                       |                       2                       |
	-- State  |   1   |   2   |   3   |   4   |   5   |   6   |   1   |   2   |   3   |   4   |   5   |   6   |
	-- Period | 1 | 2 | 1 | 2 | 1 | 2 | 1 | 2 | 1 | 2 | 1 | 2 | 1 | 2 | 1 | 2 | 1 | 2 | 1 | 2 | 1 | 2 | 1 | 2 |
	--        |    _______                 _______                 _______                 ________           |
	-- ALE    |___/       \_______________/       \_______________/       \_______________/        \__________|
	--        |    ___________             __________              __________              __________         |
	-- PSEN   |___/           \___________/          \____________/          \____________/          \________|
	--        |_______________________________________________________________________________________________|
	-- RD     |                                                                                               |
	--		  |_______ _______________________ _______________________ _______________________ _______________|
	-- P2	  |__PCH__X__PCH__________________X__PCH__________________X__PCH__________________X__PCH__________|
	--		  |___     _______ ___________     _______ ___________     _______ ___________     _______ _______|
	-- P0	 Inst_X---X__PCL__X__Inst_____X---X__PCL__X__Inst_____X---X__PCL__X__Inst_____X---X__PCL__X__Inst_|
	--        |  ^                        ^
	--        |  ^- read opcode           ^- read 2nd byte
	--
	------------------------------------
	-- Timing (B) with a MOVX
	-- Cycle  |                       1                       |                       2                       |
	-- State  |   1   |   2   |   3   |   4   |   5   |   6   |   1   |   2   |   3   |   4   |   5   |   6   |
	-- Period | 1 | 2 | 1 | 2 | 1 | 2 | 1 | 2 | 1 | 2 | 1 | 2 | 1 | 2 | 1 | 2 | 1 | 2 | 1 | 2 | 1 | 2 | 1 | 2 |
	--        |    _______                 _______                                         ________           |
	-- ALE    |___/       \_______________/       \_______________________________________/        \__________|
	--        |    ___________             __________________________________________________________         |
	-- PSEN   |___/           \___________/                                                          \________|
	--        |_______________________________________________                         _______________________|
	-- RD     |                                               \_______________________/                       |
	--		  |_______ _______________________ _______________________________________________ _______________|
	-- P2	  |__PCH__X__PCH__________________X__DPH or P2____________________________________X__PCH__________|
	--		  |___     _______ ___________     _______         _______________________         _______ _______|
	-- P0	 Inst_X---X__PCL__X__Inst_____X---X__A7:0_X-------X__Data_________________X-------X__PCL__X__Inst_|
	--        |  ^                        ^                                           ^
	--        |  ^- read opcode           ^- read 2nd byte                            ^- read data
	--
	------------------------------------
	-- constroller machine cycle/ state/ period counter
	state<= stateii;
	state_process: process(clk, rst)
	begin
		if rst= '1' then
			stateii<= C1S1P2; -- startup with NOP!
			init<=    '1';
			--stateii<= C1S5P1; -- startup with NOP!
		elsif falling_edge(clk) then
			case stateii is
				-- machine cycle #1
				when C1S1P1=> stateii<= C1S1P2; when C1S1P2=> stateii<= C1S2P1;
				when C1S2P1=> stateii<= C1S2P2; when C1S2P2=> stateii<= C1S3P1;
				when C1S3P1=> stateii<= C1S3P2; when C1S3P2=> stateii<= C1S4P1;
				when C1S4P1=> stateii<= C1S4P2; when C1S4P2=> stateii<= C1S5P1;
				when C1S5P1=> stateii<= C1S5P2; when C1S5P2=> stateii<= C1S6P1;
				when C1S6P1=> stateii<= C1S6P2;
				when C1S6P2=> if mcode.cycles= 1 then stateii<= C1S1P1; else stateii<= C2S1P1; end if; init<= '0';
				-- machine cycle #2
				when C2S1P1=> stateii<= C2S1P2; when C2S1P2=> stateii<= C2S2P1;
				when C2S2P1=> stateii<= C2S2P2; when C2S2P2=> stateii<= C2S3P1;
				when C2S3P1=> stateii<= C2S3P2; when C2S3P2=> stateii<= C2S4P1;
				when C2S4P1=> stateii<= C2S4P2; when C2S4P2=> stateii<= C2S5P1;
				when C2S5P1=> stateii<= C2S5P2; when C2S5P2=> stateii<= C2S6P1;
				when C2S6P1=> stateii<= C2S6P2;
				when C2S6P2=> if mcode.cycles= 2 then stateii<= C1S1P1; else stateii<= C3S1P1; end if;
				--
				when others=> assert(false) report "machine cycle #4 exceeded" severity failure;
			end case;
		end if;
	end process;
	
	------------------------------------
	-- opcode fetch
	opcode1<= opcode1ii;
	opcode_process: process(clk, rst)
	begin
		if rst= '1' then
			opcode1ii<= (others=> '0'); -- startup with NOP!
			opcode2ii<= (others=> '-');
			opcode3ii<= (others=> '-');
		elsif falling_edge(clk) then
			if init= '0' then
				case stateii is
					when C1S1P1=> opcode1ii<= datai; opcode2ii<= (others=> '-'); opcode3ii<= (others=> '-');
					when C1S4P1=> opcode2ii<= datai; 
					when C2S1P1=> opcode3ii<= datai; 
					when others=> null;
				end case;
			end if;
		end if;
	end process;
	
	------------------------------------
	-- alu 
	alu_process: process(stateii, pc, opcodeii, mcode, 
		dst1s_r, dstfs_r,
		dst1c_r, dstfc_r,
		tmp, tmp_r,
		src2, src2_r,
		reg_drd,
		alu_dst1, alu_dstf)
		
		procedure aluind(adx1: te51addrT; adx2: te51addrT) is
		begin
			if adx1= te51addr_aRi or adx2= te51addr_aRi then
				tmp(7 downto 0)<= reg_drd; reg_adx<= "0000000" & opcode1ii(0); reg_radx<= '1';
			else
				if adx1= te51addr_call then
					alu_src1<= reg_sp;
					alu_src2<= constant_1;
					alu_op  <= te51op_sub;
					reg_adx <= alu_dst1;
					reg_dwr <= pc(7 downto 0);
					reg_wren<= '1';
				elsif adx1= te51addr_ret then
					alu_src1<= reg_sp;
					alu_src2<= constant_1;
					alu_op  <= te51op_add;
					reg_adx <= alu_dst1;
					tmp(7 downto 0)<= reg_drd;
				else
					case adx2 is
						when te51addr_pc_rel2=>
						alu_src1       <= pc(7 downto 0);
						alu_src2       <= opcode2ii;
						alu_op         <= te51op_add;
						tmp(7 downto 0)<= alu_dst1;
						
						when te51addr_pc_rel3=>
						alu_src1       <= pc(7 downto 0);
						alu_src2       <= opcode3ii;
						alu_op         <= te51op_add;
						tmp(7 downto 0)<= alu_dst1;
						
						when others=>
						null;
					end case;
				end if;
			end if;
		end procedure;
		
		impure function alusrc(addr: te51addrT) return STD_LOGIC_VECTOR is
			variable src: STD_LOGIC_VECTOR(7 downto 0);
		begin
			case addr is
				when te51addr_A      => src:= reg_drd; reg_adx<= te51alias_A;
				when te51addr_datain => src:= (others=> '-'); reg_adx<= te51alias_A;
				when te51addr_DPL    => src:= reg_drd; reg_adx<= te51alias_DPL;
				when te51addr_DPH    => src:= reg_drd; reg_adx<= te51alias_DPH;
				when te51addr_SP     => src:= reg_drd; reg_adx<= te51alias_SP;
				when te51addr_aSP    => src:= reg_drd; reg_adx<= reg_SP;
				when te51addr_direct => src:= reg_drd; reg_adx<= opcode2ii;
				when te51addr_direct3=> src:= reg_drd; reg_adx<= opcode3ii;
				when te51addr_bit    => src:= reg_drd; reg_adx<= opcode2ii; reg_badx<= '1';
				when te51addr_Rn     => src:= reg_drd; reg_adx<= "00000" & opcode1ii(2 downto 0); reg_radx<= '1';
				when te51addr_aRi    => src:= reg_drd; reg_adx<= tmp(7 downto 0);
				when te51addr_xdata2 => src:= opcode2ii;
				when te51addr_xdata3 => src:= opcode3ii;
				when te51addr_cy16   => src:= "0000000" & dstfs_r(7);
				when te51addr_c0     => src:= constant_0;
				when te51addr_c1     => src:= constant_1;
				when te51addr_const2 => src:= constant_2;
				when te51addr_CY     => src:= reg_drd; reg_adx<= te51alias_CY; reg_badx<= '1';
				
				when te51addr_PC_rel2=> src:= dst1s_r;
				alu_src1<= pc(15 downto 8);
				alu_src2<= (others=> opcode2ii(7));
				alu_srcf<= dstfc_r;
				alu_op  <= te51op_addc;
				tmp(15 downto 8)<= alu_dst1;
				
				when te51addr_PC_rel3=> src:= dst1s_r; 
				alu_src1<= pc(15 downto 8); 
				alu_src2<= (others=> opcode3ii(7));
				alu_srcf<= dstfc_r;
				alu_op  <= te51op_addc;
				tmp(15 downto 8)<= alu_dst1;
				
				when te51addr_PC_addr11=> src:= dst1s_r;
				tmp<= pc(15 downto 11) & opcode1ii(7 downto 5) & opcode2ii(7 downto 0);
				
				when te51addr_PC_addr16=> src:= dst1s_r;
				tmp<= opcode2ii & opcode3ii;
				
				when te51addr_jc     => src:= dst1s_r; -- is this needed?
				when te51addr_jnc    => src:= dst1s_r; -- is this needed?
				when te51addr_ne     => src:= dst1s_r; -- is this needed?
				when te51addr_jeq    => src:= dst1s_r; -- is this needed?
				when te51addr_jmp    => src:= dst1s_r; -- is this needed?
				
				when te51addr_call   => src:= dst1s_r; -- is this needed?
				reg_adx <= reg_sp;
				reg_dwr <= pc(15 downto 8);
				reg_wren<= '1';			
				
				when te51addr_ret    => src:= dst1s_r; -- is this needed?
				alu_src1<= reg_sp;
				alu_src2<= constant_2;
				alu_op  <= te51op_add;
				reg_adx <= alu_dst1;
				tmp(15 downto 8)<= reg_drd;
				
				when te51addr_none   => src:= (others=> '-');
				when others          => src:= (others=> '-');
				assert(false) report "ctrl: illegal src" severity failure;
			end case;
			return src;
		end function;
		
		procedure alurslt(addr: te51addrT) is
		begin
			case addr is
				when te51addr_A      => reg_dwr<= alu_dst1; reg_wren<= '1';
				when te51addr_datain => reg_dwr<= datai;    reg_wren<= '1';
				when te51addr_DPL    => reg_dwr<= alu_dst1; reg_wren<= '1';
				when te51addr_DPH    => reg_dwr<= alu_dst1; reg_wren<= '1';
				when te51addr_SP     => reg_dwr<= alu_dst1; reg_wren<= '1';
				when te51addr_aSP    => reg_dwr<= alu_dst1; reg_wren<= '1';
				when te51addr_direct => reg_dwr<= alu_dst1; reg_wren<= '1';
				when te51addr_direct3=> reg_dwr<= alu_dst1; reg_wren<= '1';
				when te51addr_Rn     => reg_dwr<= alu_dst1; reg_wren<= '1'; reg_radx<= '1';
				when te51addr_aRi    => reg_dwr<= alu_dst1; reg_wren<= '1';
				when te51addr_jc     => pc_wren<= reg_psw(7);
				when te51addr_jnc    => pc_wren<= not(reg_psw(7));
				when te51addr_ne     => pc_wren<= dst1s_r(1);
				when te51addr_jeq    => pc_wren<= dst1s_r(0);
				when te51addr_jmp    => pc_wren<= '1';
				when te51addr_call   => pc_wren<= '1';
				when te51addr_ret    => pc_wren<= '1';
				when te51addr_CY     => reg_dwr<= alu_dst1; reg_wren<= '1'; reg_badx<= '1';
				when te51addr_bit    => reg_dwr<= alu_dst1; reg_wren<= '1';
				when te51addr_none   => null;
				when others          => null;
				assert(false) report "ctrl: illegal dst" severity failure;
			end case;
		end procedure;
		
	begin
		reg_wren<= '0';
		reg_badx<= '0';
		reg_radx<= '0';
		reg_adx <= (others=> '-'); -- METAVALUE warning?
		alu_src1<= (others=> '-');
		alu_src2<= (others=> '-');
		alu_srcf<= (others=> '-');
		alu_op  <= te51op_none;
		--		pc_wren <= '0';
		tmp     <= tmp_r;
		
		case stateii is
			-- - - - - - - - - - - - - -
			-- increment pc after opcode byte #1
			when C1S1P1=>
			pc_wren <= '0';
			alu_src1<= pc(7 downto 0);
			if init= '0' then
				alu_src2<= constant_1;
			else
				alu_src2<= constant_0;
			end if;
			alu_op  <= te51op_add;
			
			when C1S1P2=>
			alu_src1<= pc(15 downto 8);
			alu_src2<= constant_0;
			alu_srcf<= dstfc_r;
			alu_op  <= te51op_addc;
			
			-- - - - - - - - - - - - - -
			-- alu slot #1
			--	when C1S2P1=>
			--	when C1S2P2=>
			--	when C1S3P1=>
			--	when C1S3P2=>
			
			-- - - - - - - - - - - - - -
			-- increment pc after opcode byte #2
			when C1S4P1=>
			alu_src1<= pc(7 downto 0);
			if mcode.bytes> 1 then
				alu_src2<= constant_1;
			else
				alu_src2<= constant_0;
			end if;
			alu_srcf<= (others=> '0');
			alu_op  <= te51op_add;
			
			when C1S4P2=>
			alu_src1<= pc(15 downto 8);
			alu_src2<= constant_0;
			alu_srcf<= dstfc_r;
			alu_op  <= te51op_addc;
			
			-- - - - - - - - - - - - - -
			-- alu slot #2
			when C1S5P1=>
			aluind(mcode.adx21, mcode.adx22);
			
			when C1S5P2=>
			src2<= alusrc(mcode.adx22);
			
			when C1S6P1=>
			alu_src1<= alusrc(mcode.adx21);
			alu_src2<= src2_r;
			alu_srcf<= reg_psw;
			alu_op  <= mcode.op2;
			if mcode.rslt2= '1' then
				alurslt(mcode.adx21);
			end if;
			
			when C1S6P2=>
			if mcode.psw2= '1' then
				reg_adx<= te51alias_PSW;
				reg_dwr<= dstfs_r;
				reg_wren<= '1';
			end if;
			
			-- - - - - - - - - - - - - -
			-- increment pc after opcode byte #3
			when C2S1P1=>
			alu_src1<= pc(7 downto 0);
			if mcode.bytes> 2 then
				alu_src2<= constant_1;
			else
				alu_src2<= constant_0;
			end if;
			alu_srcf<= (others=> '0');
			alu_op  <= te51op_add;
			
			when C2S1P2=>
			alu_src1<= pc(15 downto 8);
			alu_src2<= constant_0;
			alu_srcf<= dstfc_r;
			alu_op  <= te51op_addc;
			
			-- - - - - - - - - - - - - -
			-- alu slot #3
			when C2S2P1=>
			aluind(mcode.adx31, mcode.adx32);
			
			when C2S2P2=>
			src2<= alusrc(mcode.adx32);
			
			when C2S3P1=>
			alu_src1<= alusrc(mcode.adx31);
			alu_src2<= src2_r;
			alu_srcf<= reg_psw;
			alu_op  <= mcode.op3;
			if mcode.rslt3= '1' then
				alurslt(mcode.adx31);
			end if;
			
			when C2S3P2=>
			if mcode.psw3= '1' then
				reg_adx <= te51alias_PSW;
				reg_dwr <= dstfs_r;
				reg_wren<= '1';
			end if;
			
			-- - - - - - - - - - - - - -
			-- alu slot #4
			--	when C2S5P1
			--	when C2S5P2
			--	when C2S6P1
			--	when C2S6P2
			
			when others=> null;
		end case;
	end process;
	
	aluregs_process: process(clk, rst)
	begin
		if rst= '1' then
			src2_r <= (others=> '-');
			dst1s_r<= (others=> '-');
			dstfs_r<= (others=> '0');
			dst1c_r<= (others=> '-');
			dstfc_r<= (others=> '0');
			pc     <= (others=> '0');
			--			pc_wren1<= '0';
			--			pc_wren2<= '0';
			--			pc_wren3<= '0';
		elsif falling_edge(clk) then
			dst1c_r<= alu_dst1;
			dstfc_r<= alu_dstf;
			
			--			pc_wren1<= pc_wren;
			--			pc_wren2<= pc_wren1;
			--			pc_wren3<= pc_wren2;
			--			pc_wren4<= pc_wren3;
			
			case stateii is
				when C1S1P1 | C1S4P1 | C2S1P1=> -- inc pc(7:0)
				pc(7 downto 0)<= alu_dst1;
				
				when C1S1P2 | C1S4P2 | C2S1P2=> -- inc pc(15:8)
				pc(15 downto 8)<= alu_dst1;
				
				when C2S4P2=>                   -- program branching
				if pc_wren= '1' then
					pc<= tmp_r;
				end if;
				
				when C1S2P1 | C1S5P1 | C2S2P1=> -- aluind
				tmp_r<= tmp;
				
				when C1S2P2 | C1S5P2 | C2S2P2=> -- alu src #2
				tmp_r<= tmp;
				src2_r<= src2;
				
				when C1S3P1 | C1S6P1 | C2S3P1=> -- alu slot
				dst1s_r<= alu_dst1;
				dstfs_r<= alu_dstf;
				--				if pc_wren= '1' then
				--					pc<= tmp_r;
				--				end if;
				
				when others=>
				null;
			end case;
		end if;
	end process;
	
	------------------------------------
	-- ale, psen, rd, wr timing
	datao<= reg_a;
	extsig_process: process(stateii, mcode, reg_dptr)
		variable mxr:   STD_LOGIC;
		variable mxnr:  STD_LOGIC;
		variable mxw:   STD_LOGIC;
		variable mxnw:  STD_LOGIC;
		variable mxrw:  STD_LOGIC;
		variable mxnrw: STD_LOGIC;
		variable mxa:   STD_LOGIC_VECTOR(15 downto 0);
	begin
		mxr:=  mcode.movxrd;                 mxnr:=  not(mxr);
		mxw:=  mcode.movxwr;                 mxnw:=  not(mxw);
		mxrw:= mcode.movxrd or mcode.movxwr; mxnrw:= not(mxrw);
		if mxrw= '0' then mxa:= pc;
		else              mxa:= reg_dptr; end if;
		
		case stateii is
			when C1S1P1=> ale<= '0';   psen<= '0';  rd<= '1';  wr<= '1';  addr<= pc;  p0adx<= '0';   p0oe<= '0';
			when C1S1P2=> ale<= '1';   psen<= '1';  rd<= '1';  wr<= '1';  addr<= pc;  p0adx<= '0';   p0oe<= '0';
			when C1S2P1=> ale<= '1';   psen<= '1';  rd<= '1';  wr<= '1';  addr<= pc;  p0adx<= '1';   p0oe<= '0';
			when C1S2P2=> ale<= '0';   psen<= '1';  rd<= '1';  wr<= '1';  addr<= pc;  p0adx<= '1';   p0oe<= '0';
			when C1S3P1=> ale<= '0';   psen<= init; rd<= '1';  wr<= '1';  addr<= pc;  p0adx<= '0';   p0oe<= '0';
			when C1S3P2=> ale<= '0';   psen<= init; rd<= '1';  wr<= '1';  addr<= pc;  p0adx<= '0';   p0oe<= '0';
			when C1S4P1=> ale<= '0';   psen<= init; rd<= '1';  wr<= '1';  addr<= pc;  p0adx<= '0';   p0oe<= '0';
			when C1S4P2=> ale<= '1';   psen<= '1';  rd<= '1';  wr<= '1';  addr<= pc;  p0adx<= '0';   p0oe<= '0';
			when C1S5P1=> ale<= '1';   psen<= '1';  rd<= '1';  wr<= '1';  addr<= mxa; p0adx<= '1';   p0oe<= '0';
			when C1S5P2=> ale<= '0';   psen<= '1';  rd<= '1';  wr<= '1';  addr<= mxa; p0adx<= '1';   p0oe<= '0';
			when C1S6P1=> ale<= '0';   psen<= mxrw; rd<= '1';  wr<= '1';  addr<= mxa; p0adx<= '0';   p0oe<= mxw;
			when C1S6P2=> ale<= '0';   psen<= mxrw; rd<= mxnr; wr<= mxnw; addr<= mxa; p0adx<= '0';   p0oe<= mxw;
			--
			when C2S1P1=> ale<= '0';   psen<= mxrw; rd<= mxnr; wr<= mxnw; addr<= mxa; p0adx<= '0';   p0oe<= mxw;
			when C2S1P2=> ale<= mxnrw; psen<= '1';  rd<= mxnr; wr<= mxnw; addr<= mxa; p0adx<= '0';   p0oe<= mxw;
			when C2S2P1=> ale<= mxnrw; psen<= '1';  rd<= mxnr; wr<= mxnw; addr<= mxa; p0adx<= mxnrw; p0oe<= mxw;
			when C2S2P2=> ale<= '0';   psen<= '1';  rd<= mxnr; wr<= mxnw; addr<= mxa; p0adx<= mxnrw; p0oe<= mxw;
			when C2S3P1=> ale<= '0';   psen<= mxrw; rd<= mxnr; wr<= mxnw; addr<= mxa; p0adx<= '0';   p0oe<= mxw;
			when C2S3P2=> ale<= '0';   psen<= mxrw; rd<= '1';  wr<= '1';  addr<= mxa; p0adx<= '0';   p0oe<= mxw;
			when C2S4P1=> ale<= '0';   psen<= mxrw; rd<= '1';  wr<= '1';  addr<= mxa; p0adx<= '0';   p0oe<= mxw;
			when C2S4P2=> ale<= '1';   psen<= '1';  rd<= '1';  wr<= '1';  addr<= mxa; p0adx<= '0';   p0oe<= '0';
			when C2S5P1=> ale<= '1';   psen<= '1';  rd<= '1';  wr<= '1';  addr<= pc;  p0adx<= '1';   p0oe<= '0';
			when C2S5P2=> ale<= '0';   psen<= '1';  rd<= '1';  wr<= '1';  addr<= pc;  p0adx<= '1';   p0oe<= '0';
			when C2S6P1=> ale<= '0';   psen<= '0';  rd<= '1';  wr<= '1';  addr<= pc;  p0adx<= '0';   p0oe<= '0';
			when C2S6P2=> ale<= '0';   psen<= '0';  rd<= '1';  wr<= '1';  addr<= pc;  p0adx<= '0';   p0oe<= '0';
			--
			when others=> null;
		end case;
	end process;
	
end BHV;

library IEEE;
use IEEE.std_logic_1164.all;
use work.te51pak.all;

entity te51c is
	port(
		clk : in STD_LOGIC;
		rst : in STD_LOGIC;
		datai : in STD_LOGIC_VECTOR (7 downto 0);
		dbga : in STD_LOGIC_VECTOR (7 downto 0);
		ale : out STD_LOGIC;
		p0adx : out STD_LOGIC;
		p0oe : out STD_LOGIC;
		psen : out STD_LOGIC;
		rd : out STD_LOGIC;
		wr : out STD_LOGIC;
		addr : out STD_LOGIC_VECTOR (15 downto 0);
		datao : out STD_LOGIC_VECTOR (7 downto 0);
		dbgv : out STD_LOGIC_VECTOR (7 downto 0)
		);
end te51c;

architecture BHV of te51c is
	
	---- Component declarations -----
	
	component te51alu
		port (
			aluop : in te51opT;
			src1 : in STD_LOGIC_VECTOR (7 downto 0);
			src2 : in STD_LOGIC_VECTOR (7 downto 0);
			srcf : in STD_LOGIC_VECTOR (7 downto 0);
			dst1 : out STD_LOGIC_VECTOR (7 downto 0);
			dstf : out STD_LOGIC_VECTOR (7 downto 0)
			);
	end component;
	component te51ctrl
		port (
			alu_dst1 : in STD_LOGIC_VECTOR (7 downto 0);
			alu_dstf : in STD_LOGIC_VECTOR (7 downto 0);
			clk : in STD_LOGIC;
			datai : in STD_LOGIC_VECTOR (7 downto 0);
			mcode : in te51mcodeT;
			reg_a : in STD_LOGIC_VECTOR (7 downto 0);
			reg_dptr : in STD_LOGIC_VECTOR (15 downto 0);
			reg_drd : in STD_LOGIC_VECTOR (7 downto 0);
			reg_psw : in STD_LOGIC_VECTOR (7 downto 0);
			reg_sp : in STD_LOGIC_VECTOR (7 downto 0);
			rst : in STD_LOGIC;
			addr : out STD_LOGIC_VECTOR (15 downto 0);
			ale : out STD_LOGIC;
			alu_op : out te51opT;
			alu_src1 : out STD_LOGIC_VECTOR (7 downto 0);
			alu_src2 : out STD_LOGIC_VECTOR (7 downto 0);
			alu_srcf : out STD_LOGIC_VECTOR (7 downto 0);
			datao : out STD_LOGIC_VECTOR (7 downto 0);
			opcode1 : out STD_LOGIC_VECTOR (7 downto 0);
			p0adx : out STD_LOGIC;
			p0oe : out STD_LOGIC;
			psen : out STD_LOGIC;
			rd : out STD_LOGIC;
			reg_adx : out STD_LOGIC_VECTOR (7 downto 0);
			reg_badx : out STD_LOGIC;
			reg_dwr : out STD_LOGIC_VECTOR (7 downto 0);
			reg_radx : out STD_LOGIC;
			reg_wren : out STD_LOGIC;
			state : out te51stateT;
			wr : out STD_LOGIC
			);
	end component;
	component te51dec
		port (
			opcode1 : in STD_LOGIC_VECTOR (7 downto 0);
			instr : out te51instrT
			);
	end component;
	component te51mcode
		port (
			clk : in STD_LOGIC;
			instr : in te51instrT;
			rst : in STD_LOGIC;
			state : in te51stateT;
			mcode : out te51mcodeT
			);
	end component;
	component te51regs
		port (
			adx : in STD_LOGIC_VECTOR (7 downto 0);
			badx : in STD_LOGIC;
			clk : in STD_LOGIC;
			dbga : in STD_LOGIC_VECTOR (7 downto 0);
			dwr : in STD_LOGIC_VECTOR (7 downto 0);
			radx : in STD_LOGIC;
			rst : in STD_LOGIC;
			wren : in STD_LOGIC;
			ao : out STD_LOGIC_VECTOR (7 downto 0);
			dbgv : out STD_LOGIC_VECTOR (7 downto 0);
			dptro : out STD_LOGIC_VECTOR (15 downto 0);
			drd : out STD_LOGIC_VECTOR (7 downto 0);
			pswo : out STD_LOGIC_VECTOR (7 downto 0);
			spo : out STD_LOGIC_VECTOR (7 downto 0)
			);
	end component;
	
	---- Signal declarations used on the diagram ----
	
	signal ALUOP : te51opT;
	signal INSTR : te51instrT;
	signal MCODE : te51mcodeT;
	signal BUS109 : STD_LOGIC_VECTOR (7 downto 0);
	signal BUS1921 : STD_LOGIC_VECTOR (7 downto 0);
	signal BUS1926 : STD_LOGIC_VECTOR (7 downto 0);
	signal BUS1927 : STD_LOGIC_VECTOR (7 downto 0);
	signal BUS2375 : STD_LOGIC_VECTOR (7 downto 0);
	signal BUS2379 : STD_LOGIC_VECTOR (7 downto 0);
	signal BUS2383 : STD_LOGIC_VECTOR (7 downto 0);
	signal BUS2387 : STD_LOGIC_VECTOR (7 downto 0);
	signal BUS2391 : STD_LOGIC_VECTOR (7 downto 0);
	signal BUS3158 : STD_LOGIC_VECTOR (15 downto 0);
	signal BUS3162 : STD_LOGIC_VECTOR (7 downto 0);
	signal BUS3166 : STD_LOGIC_VECTOR (7 downto 0);
	signal BUS3343 : STD_LOGIC_VECTOR (7 downto 0);
	signal NET1928 : STD_LOGIC;
	signal NET1942 : STD_LOGIC;
	signal NET2290 : STD_LOGIC;
	
	---- Global signals declarations -----
	signal STATE : te51stateT;
	
	
begin
	
	----  Component instantiations  ----
	
	Ualu : te51alu
	port map(
		aluop => aluop,
		dst1 => BUS2387,
		dstf => BUS2391,
		src1 => BUS2379,
		src2 => BUS2383,
		srcf => BUS2375
		);
	
	Uctrl : te51ctrl
	port map(
		addr => addr,
		ale => ale,
		alu_dst1 => BUS2387,
		alu_dstf => BUS2391,
		alu_op => aluop,
		alu_src1 => BUS2379,
		alu_src2 => BUS2383,
		alu_srcf => BUS2375,
		clk => clk,
		datai => datai,
		datao => datao,
		mcode => mcode,
		opcode1 => BUS109,
		p0adx => p0adx,
		p0oe => p0oe,
		psen => psen,
		rd => rd,
		reg_a => BUS3162,
		reg_adx => BUS1927,
		reg_badx => NET1942,
		reg_dptr => BUS3158,
		reg_drd => BUS1926,
		reg_dwr => BUS1921,
		reg_psw => BUS3166,
		reg_radx => NET2290,
		reg_sp => BUS3343,
		reg_wren => NET1928,
		rst => rst,
		state => state,
		wr => wr
		);
	
	Udec : te51dec
	port map(
		instr => instr,
		opcode1 => BUS109
		);
	
	Umcode : te51mcode
	port map(
		clk => clk,
		instr => instr,
		mcode => mcode,
		rst => rst,
		state => state
		);
	
	Uregs : te51regs
	port map(
		adx => BUS1927,
		ao => BUS3162,
		badx => NET1942,
		clk => clk,
		dbga => dbga,
		dbgv => dbgv,
		dptro => BUS3158,
		drd => BUS1926,
		dwr => BUS1921,
		pswo => BUS3166,
		radx => NET2290,
		rst => rst,
		spo => BUS3343,
		wren => NET1928
		);
	
	
end BHV;

library IEEE;
use IEEE.STD_LOGIC_1164.all;

entity te51mux is
	port(
		signal addr:    in    STD_LOGIC_VECTOR(15 downto 0);
		signal datain:  out   STD_LOGIC_VECTOR(7 downto 0);
		signal dataout: in    STD_LOGIC_VECTOR(7 downto 0);
		signal p0adx:   in    STD_LOGIC;
		signal p0oe:    in    STD_LOGIC;
		signal rd:      in    STD_LOGIC;
		signal wr:      in    STD_LOGIC;
		--
		signal port0:   inout STD_LOGIC_VECTOR(7 downto 0);
		signal port1:   out   STD_LOGIC_VECTOR(7 downto 0);
		signal port2:   out   STD_LOGIC_VECTOR(7 downto 0);
		signal port3:   out   STD_LOGIC_VECTOR(7 downto 0)
		);
end te51mux;

--------------------------------------------------------------------------------
architecture bhv of te51mux is
begin
	port0<= addr(7 downto 0) when p0adx= '1'
	else    dataout          when p0oe=  '1'
	else    (others=> 'Z');
	datain<= port0;
	
	port1<= (others=> 'Z');
	
	port2<= addr(15 downto 8);
	
	port3(7)<= 'H' when rd= '1' else '0';
	port3(6)<= 'H' when wr= '1' else '0';
	port3(5 downto 0)<= "HHHHHH";
	
end bhv;

library IEEE;
use IEEE.std_logic_1164.all;

entity te51d is
	port(
		clk : in STD_LOGIC;
		ea : in STD_LOGIC;
		rst : in STD_LOGIC;
		dbga : in STD_LOGIC_VECTOR (7 downto 0);
		ale : out STD_LOGIC;
		psen : out STD_LOGIC;
		dbgv : out STD_LOGIC_VECTOR (7 downto 0);
		port1 : out STD_LOGIC_VECTOR (7 downto 0);
		port2 : out STD_LOGIC_VECTOR (7 downto 0);
		port3 : out STD_LOGIC_VECTOR (7 downto 0);
		port0 : inout STD_LOGIC_VECTOR (7 downto 0)
		);
end te51d;

architecture BHV of te51d is
	
	---- Component declarations -----
	
	component te51c
		port (
			clk : in STD_LOGIC;
			datai : in STD_LOGIC_VECTOR (7 downto 0);
			dbga : in STD_LOGIC_VECTOR (7 downto 0);
			rst : in STD_LOGIC;
			addr : out STD_LOGIC_VECTOR (15 downto 0);
			ale : out STD_LOGIC;
			datao : out STD_LOGIC_VECTOR (7 downto 0);
			dbgv : out STD_LOGIC_VECTOR (7 downto 0);
			p0adx : out STD_LOGIC;
			p0oe : out STD_LOGIC;
			psen : out STD_LOGIC;
			rd : out STD_LOGIC;
			wr : out STD_LOGIC
			);
	end component;
	component te51mux
		port (
			addr : in STD_LOGIC_VECTOR (15 downto 0);
			dataout : in STD_LOGIC_VECTOR (7 downto 0);
			p0adx : in STD_LOGIC;
			p0oe : in STD_LOGIC;
			rd : in STD_LOGIC;
			wr : in STD_LOGIC;
			datain : out STD_LOGIC_VECTOR (7 downto 0);
			port1 : out STD_LOGIC_VECTOR (7 downto 0);
			port2 : out STD_LOGIC_VECTOR (7 downto 0);
			port3 : out STD_LOGIC_VECTOR (7 downto 0);
			port0 : inout STD_LOGIC_VECTOR (7 downto 0)
			);
	end component;
	
	---- Signal declarations used on the diagram ----
	
	signal BUS240 : STD_LOGIC_VECTOR (15 downto 0);
	signal BUS248 : STD_LOGIC_VECTOR (7 downto 0);
	signal BUS256 : STD_LOGIC_VECTOR (7 downto 0);
	signal NET285 : STD_LOGIC;
	signal NET289 : STD_LOGIC;
	signal NET293 : STD_LOGIC;
	signal NET297 : STD_LOGIC;
	
begin
	
	----  Component instantiations  ----
	
	Umux : te51mux
	port map(
		addr => BUS240,
		datain => BUS248,
		dataout => BUS256,
		p0adx => NET285,
		p0oe => NET289,
		port0 => port0,
		port1 => port1,
		port2 => port2,
		port3 => port3,
		rd => NET293,
		wr => NET297
		);
	
	Ute51c : te51c
	port map(
		addr => BUS240,
		ale => ale,
		clk => clk,
		datai => BUS248,
		datao => BUS256,
		dbga => dbga,
		dbgv => dbgv,
		p0adx => NET285,
		p0oe => NET289,
		psen => psen,
		rd => NET293,
		rst => rst,
		wr => NET297
		);
	
	
end BHV;

library IEEE;
use IEEE.std_logic_1164.all;

entity te51 is
	port(
		clk : in STD_LOGIC;
		ea : in STD_LOGIC;
		rst : in STD_LOGIC;
		ale : out STD_LOGIC;
		psen : out STD_LOGIC;
		port1 : out STD_LOGIC_VECTOR (7 downto 0);
		port2 : out STD_LOGIC_VECTOR (7 downto 0);
		port3 : out STD_LOGIC_VECTOR (7 downto 0);
		port0 : inout STD_LOGIC_VECTOR (7 downto 0)
		);
end te51;

architecture BHV of te51 is
	
	---- Component declarations -----
	
	component te51d
		port (
			clk : in STD_LOGIC;
			dbga : in STD_LOGIC_VECTOR (7 downto 0);
			ea : in STD_LOGIC;
			rst : in STD_LOGIC;
			ale : out STD_LOGIC;
			dbgv : out STD_LOGIC_VECTOR (7 downto 0);
			port1 : out STD_LOGIC_VECTOR (7 downto 0);
			port2 : out STD_LOGIC_VECTOR (7 downto 0);
			port3 : out STD_LOGIC_VECTOR (7 downto 0);
			psen : out STD_LOGIC;
			port0 : inout STD_LOGIC_VECTOR (7 downto 0)
			);
	end component;
	
	----     Constants     -----
	constant DANGLING_INPUT_CONSTANT : STD_LOGIC := 'Z';
	
	---- Declaration for Dangling input ----
	signal Dangling_Input_Signal : STD_LOGIC;
	
begin
	
	----  Component instantiations  ----
	
	Ute51d : te51d
	port map(
		dbga(0) => Dangling_Input_Signal,
		dbga(1) => Dangling_Input_Signal,
		dbga(2) => Dangling_Input_Signal,
		dbga(3) => Dangling_Input_Signal,
		dbga(4) => Dangling_Input_Signal,
		dbga(5) => Dangling_Input_Signal,
		dbga(6) => Dangling_Input_Signal,
		dbga(7) => Dangling_Input_Signal,
		ale => ale,
		clk => clk,
		ea => ea,
		port0 => port0,
		port1 => port1,
		port2 => port2,
		port3 => port3,
		psen => psen,
		rst => rst
		);
	
	
	---- Dangling input signal assignment ----
	
	Dangling_Input_Signal <= DANGLING_INPUT_CONSTANT;
	
end BHV;

--------------------------------------------------------------------------------
-- end of file
